import json
import logging
import threading

import math

import spread_core.mqtt as mqtt
import spread_core.protocols.dali as dali
from spread_core.bam.dali import bindings
from spread_core.bam.dali.providers.providers import DaliProvider, DaliFunit, Groups, Validity, Discovery, BindingDevice
from spread_core.bam.dali.tuner import TunerLimited
from spread_core.bam.multi_command import MultiCommand

""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                                  DALI 1                                """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

DISCOVERY_DELAY = 0.5


class DaliLight(DaliProvider):
    _dali_type = 1

    def __init__(self, data):
        super(DaliLight, self).__init__(data)
        self._fadeTime = 2
        self._fadeRate = 0
        self._level_timer = None
        self.is_unaddress_tunned = False
        self._tuner = None
        self.level_cmd_count = 0
        self._last_level = 0

    @property
    def fadeTime(self): return self._fadeTime

    @property
    def fadeRate(self): return self._fadeRate

    @fadeTime.setter
    def fadeTime(self, value):
        self._fadeTime = value

        if self.is_device:
            for _group in self.manager.groups_of(self.id):
                max_ft = 0
                if isinstance(_group, DaliLight) and _group.is_group:
                    for p_id in _group.get_group_members():
                        max_ft = max(max_ft, _group.manager.get_provider(p_id).fadeTime)
                    _group.fadeTime = max_ft

            broadcast = self.manager.broadcast
            if broadcast and isinstance(broadcast, DaliLight) and broadcast.is_broadcast:
                max_ft = 0
                for p_id in broadcast.get_broadcast_members():
                    max_ft = max(max_ft, self.manager.get_provider(p_id).fadeTime)

                broadcast.fadeTime = max_ft

    @fadeRate.setter
    def fadeRate(self, value):
        self._fadeRate = value

        if self.is_device:
            for _group in self.manager.groups_of(self.id):
                max_ft = 0
                if isinstance(_group, DaliLight) and _group.is_group:
                    for p_id in _group.get_group_members():
                        max_ft = max(max_ft, _group.manager.get_provider(p_id).fadeRate)
                    _group.fadeRate = max_ft

            broadcast = self.manager.broadcast
            if broadcast and isinstance(broadcast, DaliLight) and broadcast.is_broadcast:
                max_ft = 0
                for p_id in broadcast.get_broadcast_members():
                    max_ft = max(max_ft, self.manager.get_provider(p_id).fadeRate)

                broadcast.fadeRate = max_ft

    def repair_on_fail(self) -> tuple:
        return (*super(DaliLight, self).repair_on_fail(), CurrentLevelRaw, BrightnessLevel, Tuning,)

    @property
    def tuner(self) -> TunerLimited:
        return self._tuner

    def on_survey(self):
        super(DaliLight, self).on_survey()
        self.manager.add_survey(self, dali.gear.general.QueryStatus(self.short_addr.address))
        self.manager.add_survey(self, dali.gear.general.QueryActualLevel(self.short_addr.address))

    def request(self, cmd, *args, **kwargs):
        if isinstance(cmd, (dali.gear.general.StepUp, dali.gear.general.StepDown)):
            if self.tuner and not self.tuner.is_active:
                logging.debug(f'skip {cmd}')
                return

        super(DaliLight, self).request(cmd)

        if cmd.is_send_notify:
            self.on_command_sended(cmd)

    def on_request_answer(self, funit_type, response, sig=None):
        if isinstance(response, dali.gear.general.QueryFadeTimeAndRateResponse):
            ft = None if response.fade_time is None else 'ft%s' % int(FadeTime.values[response.fade_time]*10)
            self.on_update(FadeTime, ft, sig=sig)

            fr = None if response.fade_rate is None else 'fr%s' % int(FadeRate.values[response.fade_rate]*10)
            self.on_update(FadeRate, fr, sig=sig)
        elif funit_type == CurrentLevelRaw and response.value.error:
            return
        elif isinstance(response, dali.gear.general.QueryStatusResponse):
            self.on_update(Validity, not response.error, sig=sig)
        else:
            super(DaliLight, self).on_request_answer(funit_type, response, sig)

    def check_valid(self):
        super(DaliLight, self).check_valid()
        if self.is_device:
            self.request(dali.gear.general.QueryStatus(self.short_addr.address))

    def on_group_updated(self, funit_type, value, sig=None):
        if not self.is_device:
            for _p_id in self.get_members():
                _provider = self.manager.get_provider(_p_id)
                if isinstance(_provider, DaliLight) and _provider.is_device:
                    _provider.on_update(funit_type, value, sig)

    def get_info(self):
        super(DaliLight, self).get_info()

        if self.is_device:
            if FadeTime not in self.values or FadeRate not in self.values:
                self.request(dali.gear.general.QueryFadeTimeFadeRate(self.short_addr.address))
            if MinLevelRaw not in self.values:
                self.request(dali.gear.general.QueryMinLevel(self.short_addr.address))
            if MaxLevelRaw not in self.values:
                self.request(dali.gear.general.QueryMaxLevel(self.short_addr.address))
            if DimmingCurve not in self.values:
                self.request(dali.gear.led.QueryDimmingCurve(self.short_addr.address))
            if SceneLevelsRaw not in self.values:
                self.GetSceneLevelsRaw()

            self.request(dali.gear.general.QueryActualLevel(self.short_addr.address))

            if Types not in self.values:
                self.request(dali.gear.general.QueryDeviceType(self.short_addr.address))
            if PowerOnLevelRaw not in self.values:
                self.request(dali.gear.general.QueryPowerOnLevel(self.short_addr.address))
            if SystemFailureLevelRaw not in self.values:
                self.request(dali.gear.general.QuerySystemFailureLevel(self.short_addr.address))
            if PhysicalMinLevelRaw not in self.values:
                self.request(dali.gear.general.QueryPhysicalMinimum(self.short_addr.address))
        else:
            # fadeTime меняется при изменении fadeTime e объектов
            self.on_update(FadeTime, self.get_value(FadeTime, FadeTime.index_of[1]))
            self.on_update(FadeRate, self.get_value(FadeRate, FadeRate.index_of[1]))
            self.on_update(MinLevelRaw, self.get_value(MinLevelRaw, 1))
            self.on_update(MaxLevelRaw, self.get_value(MaxLevelRaw, 254))
            self.on_update(DimmingCurve, self.get_value(DimmingCurve, DimmingCurve.index_of[0]))
            self.on_update(SceneLevelsRaw, self.get_value(SceneLevelsRaw, []))
            self.on_update(CurrentLevelRaw, self.get_value(CurrentLevelRaw, 0))
            self.on_update(PowerOnLevelRaw, self.get_value(PowerOnLevelRaw, 0))
            self.on_update(SystemFailureLevelRaw, self.get_value(SystemFailureLevelRaw, 80))
            self.on_update(PhysicalMinLevelRaw, self.get_value(PhysicalMinLevelRaw, 1))

    def on_command_prepared(self, *commands, **kwargs):
        for command in commands:
            if isinstance(command, dali.gear.general._GearCommand) and command.is_levels_dependent:
                _p_ids = self.get_members()
                for _p_id in _p_ids:
                    _provider = self.manager.get_provider(_p_id)
                    if isinstance(_provider, DaliLight):
                        _provider.stop_level_timer()
                        _provider.level_cmd_count += 1

        super(DaliLight, self).on_command_prepared(*commands, **kwargs)

    def stop_level_timer(self):
        if self._level_timer:
            self._level_timer.cancel()

    def query_level(self, sig):
        if self.get_value(Discovery, False) is True:
            return

        if self.is_device:
            self.level_cmd_count -= 1

            if self.level_cmd_count < 1:
                self.level_cmd_count = 0
                cmd = dali.gear.general.QueryActualLevel(self.short_addr)
                cmd.sig = sig
                self.stop_level_timer()
                self._level_timer = threading.Timer(1+self.fadeTime, self.on_command_prepared, args=[cmd])
                self._level_timer.setName('Light.query_level Timer')
                self._level_timer.start()
        else:
            for _p_id in self.get_members():
                _provider = self.manager.get_provider(_p_id)
                if isinstance(_provider, DaliLight):
                    if _provider.is_device:
                        _provider.query_level(None)
                    else:
                        logging.warning(f'Provider({_provider.id}) is participant of {self.id}')

    def on_update(self, funit_type, value, sig=None):
        super(DaliLight, self).on_update(funit_type, value, sig)

        if funit_type in [CurrentLevelRaw, DimmingCurve]:
            self.GetBrightnessLevel(sig)
        elif funit_type == FadeTime:
            try:
                self.fadeTime = FadeTime.values[FadeTime.index_of.index(value)]
            except:
                pass
        elif funit_type == FadeRate:
            try:
                self.fadeRate = FadeRate.values[FadeRate.index_of.index(value)]
            except:
                pass

    def on_command_sended(self, cmd):
        if cmd.is_levels_dependent:
            _p_ids = self.get_members()
            if not self.is_device:
                _p_ids.append(self.id)

            for _p_id in _p_ids:
                _provider = self.manager.get_provider(_p_id)
                if isinstance(_provider, DaliLight):
                    value = None
                    cl = _provider.get_value(CurrentLevelRaw, 0)

                    if isinstance(cmd, dali.gear.general.DAPC):
                        value = cmd.frame.pack[1]
                    elif isinstance(cmd, dali.gear.general.Off):
                        value = 0
                    elif isinstance(cmd, dali.gear.general.RecallMinLevel):
                        value = 1
                    elif isinstance(cmd, dali.gear.general.RecallMaxLevel):
                        value = 254
                    elif isinstance(cmd, dali.gear.general.GoToLastActiveLevel):
                        value = _provider._last_level
                    elif isinstance(cmd, dali.gear.general.Up):
                        if cl > 0:
                            value = _provider.get_value(CurrentLevelRaw, 0) + max(1, round(self.fadeRate/5))
                    elif isinstance(cmd, dali.gear.general.Down):
                        if cl > 0:
                            value = max(1, _provider.get_value(CurrentLevelRaw, 0) - max(1, round(self.fadeRate/5)))
                    elif isinstance(cmd, dali.gear.general.StepUp):
                        if cl > 0:
                            value = _provider.get_value(CurrentLevelRaw, 0) + 1
                    elif isinstance(cmd, dali.gear.general.OnAndStepUp):
                        value = _provider.get_value(CurrentLevelRaw, 0) + 1
                    elif isinstance(cmd, dali.gear.general.StepDown):
                        if cl > 0:
                            value = _provider.get_value(CurrentLevelRaw, 0) - 1
                    elif isinstance(cmd, dali.gear.general.StepDownAndOff):
                        value = _provider.get_value(CurrentLevelRaw, 0) - 1
                    elif isinstance(cmd, dali.gear.general.GoToScene):
                        scenes = _provider.get_value(SceneLevelsRaw, None)
                        scene_id = cmd.frame.pack[1] - 0x10
                        if scenes is not None and scene_id in scenes and scenes[scene_id] in range(0, 254):
                            value = scenes[scene_id]

                    if value is not None:
                        if _provider.is_valid:
                            if value > 0:
                                value = max(_provider.get_value(MinLevelRaw, 0), value)
                                value = min(_provider.get_value(MaxLevelRaw, 254), value)
                                _provider._last_level = value
                            _provider.on_update(CurrentLevelRaw, value)
            if self.tuner:
                self.tuner.check_state(cmd)

            self.query_level(cmd.sig)

        super(DaliLight, self).on_command_sended(cmd)

    def prs2raw(self, p):
        if p == 0:
            return p

        p = abs(p)
        _min = self.get_value(MinLevelRaw, 1)
        _max = self.get_value(MaxLevelRaw, 254)

        if self.get_value(DimmingCurve, None) == DimmingCurve.LINEAR:
            return min(max(_min, round(254 * p / 100)), _max)
        else:
            return min(max(_min, round(((math.log10(p) + 1) * 253 / 3 + 1))), _max)

    def raw2prs(self, r):
        if self.get_value(DimmingCurve, None) == DimmingCurve.LINEAR:
            return min(100, round(100 * r / 254))
        else:
            if r == 0:
                return 0
            elif r < 1:
                return 1
            else:
                p = 10 ** ((r - 1) * 3 / 253 - 1)
                if 0 < p < 1:
                    p = 1
                return round(p)

    def read_memory_location(self, bank_id, start_address, end_address) -> bytes:
        self.check_command_for_device()
        i = self.manager.send_interface

        i.send(dali.gear.general.DTR1(bank_id))
        result = b''

        for _dtr0 in range(start_address, end_address):
            i.send(dali.gear.general.DTR0(_dtr0))
            res = i.send(dali.gear.general.ReadMemoryLocation(self.short_addr.address))
            if isinstance(res, dali.command.Response):
                result += res.value.pack
            else:
                break

        return result

    def SetShortAddress(self, new_addr, sig=None):
        if self.is_device:
            if new_addr != self.short_addr.address:
                values = None
                old_short = self.short_addr
                i = self.send
                if new_addr is None:
                    i(dali.gear.general.DTR0(0xff))
                    i(dali.gear.general.SetShortAddress(old_short))
                else:
                    new_short = dali.address.Short(new_addr)
                    res = i(dali.gear.general.QueryDeviceType(new_short))
                    if res.value.error:
                        i(dali.gear.general.DTR0((new_short.address << 1) | 1))
                        i(dali.gear.general.SetShortAddress(old_short))
                    else:
                        for _provider in self.manager.providers_by_address(new_addr):
                            if isinstance(_provider, self.__class__):
                                old_long = []
                                new_long = []

                                i(dali.gear.general.Terminate())

                                i(dali.gear.general.Initialise(broadcast=False, address=old_short.address))
                                i(dali.gear.general.Initialise(broadcast=False, address=new_short.address))

                                old_long.append(i(dali.gear.general.QueryRandomAddressH(old_short)))
                                old_long.append(i(dali.gear.general.QueryRandomAddressM(old_short)))
                                old_long.append(i(dali.gear.general.QueryRandomAddressL(old_short)))

                                new_long.append(i(dali.gear.general.QueryRandomAddressH(new_short)))
                                new_long.append(i(dali.gear.general.QueryRandomAddressM(new_short)))
                                new_long.append(i(dali.gear.general.QueryRandomAddressL(new_short)))

                                i(dali.gear.general.SetSearchAddrH(old_long[0].value.as_integer))
                                i(dali.gear.general.SetSearchAddrM(old_long[1].value.as_integer))
                                i(dali.gear.general.SetSearchAddrL(old_long[2].value.as_integer))
                                i(dali.gear.general.ProgramShortAddress(new_short.address))

                                i(dali.gear.general.SetSearchAddrH(new_long[0].value.as_integer))
                                i(dali.gear.general.SetSearchAddrM(new_long[1].value.as_integer))
                                i(dali.gear.general.SetSearchAddrL(new_long[2].value.as_integer))
                                i(dali.gear.general.ProgramShortAddress(old_short.address))

                                i(dali.gear.general.Terminate())

                                values = _provider.values.copy()

                                for funit_type, value in self.values.items():
                                    _provider.on_update(funit_type, value)

                                break

                self.SetBindingDevice(new_addr, sig=sig, values=values)
            else:
                raise dali.exceptions.CommandGenerateError('You can not give an address to {}'.format(self.binding.b_type))

    def SetDiscovery(self, value, sig=None):
        _discovery = self.get_value(Discovery, False)
        if _discovery != value:
            self.on_update(Discovery, value, sig=sig)
            if _discovery is False and value is True:
                self.set_discovery_max(sig)

    def set_discovery_max(self, sig):
        if self.get_value(Discovery) is True:
            self.on_command_prepared(dali.gear.general.RecallMaxLevel(self.short_addr))
            threading.Timer(DISCOVERY_DELAY, function=self.set_discovery_min, args=[sig]).start()
        else:
            self.reset_discovery(sig)

    def set_discovery_min(self, sig):
        if self.get_value(Discovery) is True:
            self.on_command_prepared(dali.gear.general.Off(self.short_addr))
            threading.Timer(DISCOVERY_DELAY, function=self.set_discovery_max, args=[sig]).start()
        else:
            self.reset_discovery(sig)

    def reset_discovery(self, sig):
        self.SetBrightnessLevel(self.get_value(BrightnessLevel, 0), sig=sig)

    def GetSceneLevelsRaw(self, sig=None):
        if self.is_device:
            _levels = []
            for _scene_index in range(0, 16):
                resp = self.manager.send_interface.send(dali.gear.general.QuerySceneLevel(self.short_addr, _scene_index))
                if resp is not None and resp.value is not None:# and resp.value.as_integer < 0xff:
                    _levels.append(dict(index=_scene_index, value=resp.value.as_integer))
        else:
            _levels = self.get_value(SceneLevelsRaw)

        self.on_update(SceneLevelsRaw, _levels, sig=sig)

    def SetSceneLevelsRaw(self, val, method=None, sig=None):
        if isinstance(val, str):
            val = json.loads(val)
        _val = dict()
        for item in val:
            if method == BrightnessLevel:
                item['value'] = self.prs2raw(item['value'])
            _val[item['index']] = item['value']

        c_val = self.get_value(SceneLevelsRaw, dict())
        _c_val = dict()
        for item in c_val:
            _c_val[item['index']] = item['value']

        for _i in range(0, 16):
            _v = None
            if _i in _val:
                if _i not in _c_val or _c_val[_i] != _val[_i]:
                    self.request(dali.gear.general.DTR0(_val[_i]))
                    self.manager.send_interface.send(dali.gear.general.SetScene(self.short_addr, _i))
            elif _i in _c_val:
                val.append(dict(index=_i, value=_c_val[_i]))
            else:
                val.append(dict(index=_i, value=0xff))

        self.on_update(SceneLevelsRaw, val, sig=sig)
        self.on_group_updated(SceneLevelsRaw, val, sig)

    def GetGroups(self, sig=None):
        if self.is_device:
            i = self.manager.send_interface
            _groups = []

            resp_0_7 = i.send(dali.gear.general.QueryGroupsZeroToSeven(self.short_addr))
            if resp_0_7 and resp_0_7.value is not None and not resp_0_7.value.error:
                for _i in range(0, 8):
                    _include = resp_0_7.value.as_integer & (1 << _i) == (1 << _i)
                    _groups.append(dict(index=_i, value=_include))
                    if _include:
                        self.manager.add_to_group(1, _i, self.id)

            resp_8_15 = i.send(dali.gear.general.QueryGroupsEightToFifteen(self.short_addr))
            if resp_8_15 and resp_8_15.value is not None and not resp_8_15.value.error:
                for _i in range(0, 8):
                    _include = resp_8_15.value.as_integer & (1 << _i) == (1 << _i)
                    _groups.append(dict(index=(8+_i), value=_include))
                    if _include:
                        self.manager.add_to_group(1, 8+_i, self.id)
        else:
            _groups = [{'index': self.binding.addr.group, 'value': True}] if self.is_group else []
            _groups = self.get_value(Groups, _groups)

        self.on_update(Groups, _groups, sig=sig)

    def SetGroups(self, val, sig=None):
        i = self.manager.send_interface
        if not isinstance(val, list):
            val = json.loads(val)
        _val = dict()
        for item in val:
            _val[item['index']] = item['value']

        c_val = self.get_value(Groups, dict())
        _c_val = dict()
        for item in c_val:
            _c_val[item['index']] = item['value']

        for _i, _inc in _val.items():
            if _inc is True:
                if _i not in _c_val or _c_val[_i] != _val[_i]:
                    i.send(dali.gear.general.AddToGroup(self.short_addr, _i))
                    if self.is_device:
                        self.manager.add_to_group(1, _i, self.id)
            elif _i in _c_val and _c_val[_i] is True:
                if self.is_group and self.binding.addr.group == _i:
                    continue
                i.send(dali.gear.general.RemoveFromGroup(self.short_addr, _i))
                if self.is_device:
                    self.manager.remove_from_group(1, _i, self.id)

            _c_val[_i] = _inc

        res = []
        for _i, _inc in _c_val.items():
            res.append({'index': _i, 'value': _inc})

        self.on_update(Groups, res, sig=sig)
        self.on_group_updated(Groups, val, sig)

    def SetGroupLevel(self, value, sig=None):
        try:
            value = int(value)
        except:
            raise dali.exceptions.CommandGenerateError('Value({}) must be integer'.format(value))
        else:
            group_ids = self.get_value(Groups, None)
            group_id = None
            group_size = 999999999
            if group_ids is None:
                self.GetGroups()
                group_ids = self.get_value(Groups, [])

            for item in group_ids:
                if item['value'] is True and 'index' in item:
                    _group_id = item['index']
                    if _group_id in self.manager._groups_1:
                        _group_size = len(self.manager._groups_1[_group_id])
                        if _group_size < group_size:
                            group_size = _group_size
                            group_id = _group_id

            if group_id is not None:
                _provider = None
                for _p_id, _p in self.manager.providers.items():
                    if isinstance(_p, DaliLight) and isinstance(_p.binding.addr, bindings.Group) and group_id == _p.binding.addr.group:
                        _provider = _p
                        break
                if _provider is not None:
                    _provider.SetBrightnessLevel(value, sig=sig)
                else:
                    self.set_pause_for_group(group_id)
                    group_addr = bindings.Group(group_id)
                    value_raw = self.prs2raw(value)
                    for _p_id in self.manager.get_group_members(self.dali_type, group_id):
                        _participant = self.manager.get_provider(_p_id)
                        _participant.SetBrightnessLevel(value, sig=sig, fake=True)
                    self.send(dali.gear.general.DAPC(group_addr, value_raw))
            else:
                self.SetBrightnessLevel(value, sig=sig)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        if self.is_valid is False:
            raise dali.exceptions.CommandGenerateError('Device is invalid!')

        method = kwargs.get(mqtt.const.METHOD, None)
        if method == 'RecallMinLevel':
            cmd = dali.gear.general.RecallMinLevel(self.short_addr)
        elif method == 'RecallMaxLevel':
            cmd = dali.gear.general.RecallMaxLevel(self.short_addr)
        elif method == 'Scene':
            cmd = dali.gear.general.GoToScene(self.short_addr, value)
        elif method == 'Off':
            cmd = dali.gear.general.Off(self.short_addr)
        elif method == 'GoToLastActiveLevel':
            cmd = dali.gear.general.GoToLastActiveLevel(self.short_addr)
        elif method == dali.gear.general.StepUp.__name__:
            cmd = dali.gear.general.StepUp(self.short_addr)
        elif method == dali.gear.general.StepDown.__name__:
            cmd = dali.gear.general.StepDown(self.short_addr)
        elif method == dali.gear.general.OnAndStepUp.__name__:
            cmd = dali.gear.general.OnAndStepUp(self.short_addr)
        elif method == dali.gear.general.StepDownAndOff.__name__:
            cmd = dali.gear.general.StepDownAndOff(self.short_addr)
        elif method == dali.gear.general.Up.__name__:
            cmd = dali.gear.general.Up(self.short_addr)
        elif method == dali.gear.general.Down.__name__:
            cmd = dali.gear.general.Down(self.short_addr)
        else:
            cmd = dali.gear.general.DAPC(self.short_addr, self.prs2raw(value))
        cmd.sig = sig
        if self.is_device:
            if self.is_unaddress_tunned:
                groups = self.get_value(Groups, [])
                for item in groups:
                    if item['value'] is True:
                        group_id = item['index']
                        self.set_pause_for_group(group_id)
                self.set_pause_for_group(-1)
            elif self.tuner:
                self.tuner.set_pause(2+self.fadeTime)
        else:
            if self.tuner:
                self.tuner.set_pause(2+self.fadeTime)
        self.on_command_prepared(cmd, fake=True)
        if kwargs.get('fake', False) is False:
            self.send(cmd)
        self.on_command_sended(cmd)

    def GetBrightnessLevel(self, sig=None):
        _level = self.get_value(CurrentLevelRaw, None)
        if _level is not None:
            _level = self.raw2prs(_level)
        self.on_update(BrightnessLevel, _level, sig=sig)

    def set_pause_for_group(self, group_id):
        for p_id, _provider in self.manager.providers.items():
            if isinstance(_provider, DaliDimmer) and _provider.tuner:
                if ((_provider.is_group and _provider.short_addr.group == group_id)
                        or (_provider.is_broadcast and group_id == -1)):
                    _provider.tuner.set_pause(2+self.fadeTime)

    def serialize(self) -> dict:
        d = super(DaliLight, self).serialize()
        if d:
            for funit_type in [Types, PhysicalMinLevelRaw, MinLevelRaw, MaxLevelRaw, PowerOnLevelRaw, Groups,
                               SystemFailureLevelRaw, FadeTime, FadeRate, DimmingCurve, SceneLevelsRaw]:
                value = self.get_value(funit_type)
                if value is not None:
                    key = str(funit_type)[0].lower() + str(funit_type)[1:]
                    d[key] = value
        return d


""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                                 DIMMER                                 """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""


class DaliDimmer(DaliLight):
    def __init__(self, data):
        super(DaliDimmer, self).__init__(data)
        self.step_cmd = None

    def SetTuning(self, value, speed=None, sig=None):
        _c_val = self.get_value(Tuning, 0)
        if _c_val * value <= 0:
            speed = 0.1 if speed is None else speed
            if self._tuner is None:
                self._tuner = TunerLimited(self)
            self.on_update(Tuning, value, sig=sig)
            if value == 0:
                self._tuner.stop()
            elif value > 0:
                self._tuner.set_direction(dali.gear.general.StepUp(self.short_addr), speed)
            else:  # value < 0
                self._tuner.set_direction(dali.gear.general.StepDown(self.short_addr), speed)
        elif self._tuner:
            self._tuner.speed = speed


""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                              COLOR CONTROL                             """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""


class DaliTunableWhite(DaliDimmer):
    def __init__(self, *args, **kwargs):
        super(DaliTunableWhite, self).__init__(*args, **kwargs)
        self.temperature_timer = None

    def repair_on_fail(self) -> tuple:
        return (*super(DaliLight, self).repair_on_fail(), Temperature,)

    def get_info(self):
        super(DaliTunableWhite, self).get_info()
        if self.is_device:
            self.GetTemperature()
            if Coolest not in self.values:
                self.GetCoolest()
            if Warmest not in self.values:
                self.GetWarmest()
        else:
            self.on_update(Temperature, self.get_value(Temperature, 5500))
            self.on_update(Coolest, self.get_value(Coolest, 2000))
            self.on_update(Warmest, self.get_value(Warmest, 9000))

    def _get_color_value(self, funit_type, value_type: int, sig=None):
        res = None
        try:
            self.manager.send_interface.send(dali.gear.general.DTR0(value_type))
            res = self.manager.send_interface.send(dali.gear.colour_control.QueryColorValue(self.short_addr)).value.as_integer
            if res != 0xff:
                b0 = self.manager.send_interface.send(dali.gear.general.QueryContentDTR0(self.short_addr)).value.as_integer
                b1 = self.manager.send_interface.send(dali.gear.general.QueryContentDTR1(self.short_addr)).value.as_integer
                res = int(1000000 / int.from_bytes([b0, b1], 'little'))
        except:
            pass
        finally:
            self.on_update(funit_type, res, sig=sig)

    def _set_color_limit(self, value_type: int, value: int):
        if not 1000 <= value <= 20000:
            raise ValueError("temperature must be in the range 1000..20000")
        mirek = int(1000000 / value)
        self._params = [int(b) for b in mirek.to_bytes(2, 'little')] + [value_type]
        self.manager.send_interface.send(dali.gear.colour_control.StoreColourTemperatureLimit(self.short_addr))

    def SetTemperature(self, temperature=1, sig=None):
        if not 1000 <= temperature <= 20000:
            raise ValueError("temperature must be in the range 1000..20000")
        try:
            mirek = int(1000000 / temperature)
            bts = [int(b) for b in mirek.to_bytes(2, 'little')]
            self.manager.send_interface.send(dali.gear.general.DTR0(bts[0]))
            self.manager.send_interface.send(dali.gear.general.DTR1(bts[1]))
            self.manager.send_interface.send(dali.gear.colour_control.SetTemperature(self.short_addr))
            self.manager.send_interface.send(dali.gear.colour_control.Activate(self.short_addr))
        except BaseException as ex:
            logging.error(ex)
        else:
            self.query_temperature()

    def query_temperature(self):
        if self.temperature_timer:
            self.temperature_timer.cancel()
        mc = MultiCommand(self.GetTemperature)
        self.temperature_timer = threading.Timer(interval=self.fadeTime, function=self.manager.add_command, args=[self, mc])
        self.temperature_timer.start()

    def GetTemperature(self, sig=None):
        self._get_color_value(Temperature, dali.gear.colour_control.QueryColorValue.TEMPERATURE, sig=sig)

    def SetCoolest(self, sig=None, value=0):
        try:
            self._set_color_limit(dali.gear.colour_control.StoreColourTemperatureLimit.COOLEST, value)
        except BaseException as ex:
            logging.error(ex)
        else:
            self.GetCoolest(sig)

    def GetCoolest(self, sig=None):
        self._get_color_value(Coolest, dali.gear.colour_control.QueryColorValue.COOLEST, sig=sig)

    def SetWarmest(self, sig=None, value=0):
        try:
            self._set_color_limit(dali.gear.colour_control.StoreColourTemperatureLimit.WARMEST, value)
        except BaseException as ex:
            logging.error(ex)
        else:
            self.GetWarmest(sig)

    def GetWarmest(self, sig=None):
        self._get_color_value(Warmest, dali.gear.colour_control.QueryColorValue.WARMEST, sig=sig)


""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                                 SWITCH                                 """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""


class DaliRelay(DaliLight):
    pass


class DaliLightFunit(DaliFunit):
    entity_type = DaliLight
    _dtr0 = dali.gear.general.DTR0
    _dtr1 = dali.gear.general.DTR1
    _dtr2 = dali.gear.general.DTR2

    @property
    def entity(self) -> entity_type: return super(DaliLightFunit, self).entity

    def _prepare_cmd(self, provider: DaliLight, cmd, *args):
        if cmd._hasparam:
            return cmd(provider.short_addr, *args)
        else:
            return cmd(provider.short_addr)


class BrightnessLevel(DaliLightFunit):
    is_control = True
    set_func = DaliLight.SetBrightnessLevel
    get_func = DaliLight.GetBrightnessLevel


class Types(DaliLightFunit):
    value_type = list
    easy_get = True


class CurrentLevelRaw(DaliLightFunit):
    is_control = True
    get_after_set = False
    set_cmd = dali.gear.general.DAPC
    get_cmd = dali.gear.general.QueryActualLevel


class PhysicalMinLevelRaw(DaliLightFunit):
    get_cmd = dali.gear.general.QueryPhysicalMinimum


class MinLevelRaw(DaliLightFunit):
    set_cmd = dali.gear.general.SetMinLevel
    get_cmd = dali.gear.general.QueryMinLevel


class MaxLevelRaw(DaliLightFunit):
    set_cmd = dali.gear.general.SetMaxLevel
    get_cmd = dali.gear.general.QueryMaxLevel


class PowerOnLevelRaw(DaliLightFunit):
    set_cmd = dali.gear.general.SetPowerOnLevel
    get_cmd = dali.gear.general.QueryPowerOnLevel


class SystemFailureLevelRaw(DaliLightFunit):
    set_cmd = dali.gear.general.SetSystemFailureLevel
    get_cmd = dali.gear.general.QuerySystemFailureLevel


class SceneLevelsRaw(DaliLightFunit):
    set_func = DaliLight.SetSceneLevelsRaw
    get_func = DaliLight.GetSceneLevelsRaw


class FadeTime(DaliLightFunit):
    values = [0, 0.7, 1, 1.4, 2, 2.8, 4, 5.7, 8, 11.3, 16, 22.6, 32, 45.3, 64, 90.5]
    index_of = ["ft0", "ft7", "ft10", "ft14", "ft20", "ft28", "ft40", "ft57", "ft80", "ft113", "ft160", "ft226",
                "ft320", "ft453", "ft640", "ft905"]
    set_cmd = dali.gear.general.SetFadeTime
    get_cmd = dali.gear.general.QueryFadeTimeFadeRate


class FadeRate(DaliLightFunit):
    values = [0, 358, 253, 179, 127, 89.4, 63.3, 44.7, 31.6, 22.4, 15.8, 11.2, 7.9, 5.6, 4, 2.8]
    index_of = ["", "fr3580", "fr2530", "fr1790", "fr1270", "fr894", "fr633", "fr447", "fr316", "fr224", "fr158",
                "fr112", "fr79", "fr56", "fr40", "fr28"]
    set_cmd = dali.gear.general.SetFadeRate
    get_cmd = dali.gear.general.QueryFadeTimeFadeRate


class DimmingCurve(DaliLightFunit):
    LOGARITHMIC = 'Logarithmic'
    LINEAR = 'Linear'
    index_of = [LOGARITHMIC, LINEAR]
    set_cmd = dali.gear.led.SelectDimmingCurve
    get_cmd = dali.gear.led.QueryDimmingCurve


class GroupLevel(DaliLightFunit):
    is_control = True
    set_func = DaliLight.SetGroupLevel


class Tuning(DaliLightFunit):
    entity_type = DaliDimmer
    set_func = entity_type.SetTuning


class Temperature(DaliFunit):
    entity_type = DaliTunableWhite
    set_func = entity_type.SetTemperature
    get_func = entity_type.GetTemperature


class Warmest(DaliFunit):
    entity_type = DaliTunableWhite
    set_func = entity_type.SetWarmest
    get_func = entity_type.GetWarmest


class Coolest(DaliFunit):
    entity_type = DaliTunableWhite
    set_func = entity_type.SetCoolest
    get_func = entity_type.GetCoolest
