import logging

import spread_core.mqtt as mqtt
import spread_core.protocols.dali as dali
from spread_core.bam.dali import bindings
from spread_core.bam.entities import Funit
from spread_core.bam.multi_command import MultiCommand
from spread_core.bam.providers import Provider
from spread_core.errors.project_errors import JocketError
from spread_core.tools import settings
from spread_core.tools.signature import CommandSignature


class DaliProvider(Provider):
    _dali_type = 0
    dali_type = property(fget=lambda self: self._dali_type)

    def __init__(self, data):
        super(DaliProvider, self).__init__(data)

        f_unit = bindings.BINDING
        _exist, _ower = settings.dump.get_dump(data['type'], data['id'], f_unit)
        data['attributes']['def_' + f_unit.lower()] = data['attributes'][f_unit.lower()]
        if _exist:
            data['attributes'][f_unit] = _ower

        f_unit = bindings.BINDING + bindings.DEVICE
        data['attributes']['def_' + bindings.DEVICE.lower()] = data['attributes'][bindings.DEVICE.lower()]
        _exist, _ower = settings.dump.get_dump(data['type'], data['id'], f_unit)
        if _exist:
            data['attributes'][bindings.DEVICE.lower()] = _ower

        f_unit = bindings.BINDING + bindings.GROUP
        data['attributes']['def_' + bindings.GROUP.lower()] = data['attributes'][bindings.GROUP.lower()]
        _exist, _ower = settings.dump.get_dump(data['type'], data['id'], f_unit)
        if _exist:
            data['attributes'][bindings.GROUP.lower()] = _ower

        self._binding = bindings.of(data['attributes'])

    def __str__(self):
        return super(DaliProvider, self).__str__() + f' {self.binding}'

    def set_value(self, key, value):
        super(DaliProvider, self).set_value(key, value)
        if key in (BindingDevice, BindingGroup, Binding):
            if (key == BindingDevice and value == self.binding.def_device) or \
                    (key == BindingGroup and value == self.binding.def_group) or \
                    (key == Binding and value == self.binding.def_binding):
                settings.dump.set_dump(self, key, settings.KILL_ITEM)
            else:
                settings.dump.set_dump(self, key, value)

    def not_repaired(self) -> tuple:
        return (*super(DaliProvider, self).not_repaired(), Binding, BindingDevice, BindingGroup,)

    @property
    def surveyable(self):
        if self.is_device:
            return super(DaliProvider, self).surveyable

        return False

    @property
    def binding(self):
        return self._binding

    @property
    def short_addr(self):
        return self.binding.addr

    @property
    def is_device(self):
        return isinstance(self.binding, bindings.BindingDevice)

    @property
    def is_group(self):
        return isinstance(self.binding, bindings.BindingGroup)

    @property
    def is_broadcast(self):
        return isinstance(self.binding, bindings.BindingBroadcast)

    @property
    def is_unaddressed(self):
        return isinstance(self.binding, bindings.BroadcastUnaddressed)

    def get_group_members(self):
        if self.is_group:
            if self.dali_type > 0:
                return self.manager.get_group_members(self.dali_type, self.short_addr.group)
        return []

    def get_broadcast_members(self):
        if self.is_broadcast:
            if self.dali_type > 0:
                return self.manager.get_all_devices(self.dali_type)
        return []

    def get_members(self) -> []:
        if self.is_group:
            return self.get_group_members()
        elif self.is_broadcast:
            return self.get_broadcast_members()

        return [self.id]

    def on_state(self, funit_type, value):
        super(DaliProvider, self).on_state(funit_type, value)

        if funit_type == Groups and self.is_device and value:
            for item in value:
                if item['value'] is True:
                    self.manager.add_to_group(self.dali_type, item['index'], self.id)

    def request(self, cmd: dali.command.Command, *args, **kwargs):
        try:
            response = self._manager.send_interface.send(cmd, *args, **kwargs)
            if cmd.is_query:
                [self.on_request_answer(funit_type, response, cmd.sig) for funit_type in cmd.funit_type]
        except BaseException as ex:
            logging.exception(ex)

    def on_request_answer(self, funit_type, response, sig=None):
        if isinstance(response, dali.command.YesNoResponse):
            self.on_update(funit_type, response.value, sig=sig)
        elif isinstance(response, dali.command.Response) and not isinstance(response.value, dali.frame.BackwardModuleError):
            funit = DaliFunit.of(funit_type)
            if issubclass(funit, DaliFunit):
                if funit.index_of:
                    value = funit.index_of[response.value.as_integer]
                elif funit.mask:
                    value = [funit.mask[i] for i in range(0, 8) if response.value.as_integer & 2**i != 0]
                else:
                    value = response.value.as_integer

                if funit.value_type:
                    if value.__class__ != funit.value_type:
                        if funit.value_type == list:
                            value = [value]
                        else:
                            value = funit.value_type(value)
            else:
                value = response.value.as_integer

            self.on_update(funit_type, value, sig=sig)
        else:
            super(DaliProvider, self).on_request_answer(funit_type, None, sig)

    def on_update(self, funit_type, value, sig=None):
        if funit_type == Validity:
            if self.is_valid and not value:
                self.set_valid(self.SUSPICIOUS)
            elif self._status != self.VALID and value:
                self.set_valid(self.VALID)
        else:
            super().on_update(funit_type, value, sig)

            if funit_type == BindingDevice and value is not None and self.short_addr.address != value:
                if isinstance(self.binding, bindings.BindingDevice):
                    try:
                        self.manager._addresses[self.short_addr.address].remove(self.id)
                        if len(self.manager._addresses[self.short_addr.address]) == 0:
                            self.manager._addresses.pop(self.short_addr.address)
                    finally:
                        self.binding.set_addr(value)
                        self.manager._addresses.setdefault(value, [])
                        self.manager._addresses[value].append(self.id)

    def on_exit(self):
        self.publish_all(True)

    def get_bindings(self):
        self.GetBinding()
        self.GetBindingDevice()
        self.GetBindingGroup()

    def get_info(self):
        # if F_SERIAL not in self.values:
        #     self.manager.add_info(self, MultiCommand(F_SERIAL, self.Serial))
        # if F_GTIN not in self.values:
        #     self.manager.add_info(self, MultiCommand(F_GTIN, self.Gtin))
        # if F_FIRMWARE not in self.values:
        #     self.manager.add_info(self, MultiCommand(F_FIRMWARE, self.FirmwareVersion))
        # if F_HARDWARE not in self.values:
        #     self.manager.add_info(self, MultiCommand(F_HARDWARE, self.HardwareVersion))
        # if F_GTIN_OEM not in self.values:
        #     self.manager.add_info(self, MultiCommand(F_GTIN_OEM, self.GtinOem))
        # if F_SERIAL_OEM not in self.values:
        #     self.manager.add_info(self, MultiCommand(F_SERIAL_OEM, self.SerialOem))
        if Discovery not in self.values:
            self.on_update(Discovery, False)
        if Groups not in self.values:
            self.GetGroups()

    def on_survey(self):
        pass

    def check_valid(self):
        super(DaliProvider, self).check_valid()
        self._status = self.SUSPICIOUS

    def set_valid(self, validity):
        if validity == self._status == self.SUSPICIOUS:
            validity = self.INVALID

        super(DaliProvider, self).set_valid(validity)

        if validity != self.SUSPICIOUS:
            for funit_type in self.values.copy():
                if funit_type not in (Binding, BindingDevice, BindingGroup):
                    self.on_update(funit_type, None if validity == self.INVALID else self.get_value(funit_type))
                    # self.values.pop(funit_type)

            # ToDo set INFO QUERIES ro survey
            if validity == self.VALID:
                self.on_command_prepared(MultiCommand(self.get_info))

    def reset(self):
        self.publish_all(invalid=True)
        self.values = {}
        self.check_valid()
        self.get_bindings()
        if self.is_valid:
            self.get_info()

    def check_command_for_device(self):
        if not self.is_device:
            raise JocketError('The GET command cannot be applied to an object that is not an DEVICE!')

    def read_memory_location(self, bank_id, start_address, end_address) -> bytes:
        raise NotImplementedError()

    def SetShortAddress(self, new_addr, sig=None):
        raise NotImplementedError()

    def SetDiscovery(self, value, sig=None):
        raise NotImplementedError()

    def GetBinding(self, sig=None):
        self.on_update(Binding, self.binding.b_type, sig=sig)

    def SetBinding(self, value, sig=None):
        if value not in bindings.ALL:
            raise dali.exceptions.CommandGenerateError('Binding({}) is not present in available values({})'.format(value, bindings.ALL))

        _c_binding = self.get_value(bindings.BINDING, bindings.DEVICE)
        if _c_binding != value:
            self._binding = bindings.of({
                bindings.BINDING.lower(): value,
                value.lower(): self.get_value((str(Binding) + value), -1),
                'def_' + bindings.BINDING.lower(): self.binding.def_binding
            })

            self.on_update(Binding, value, sig=sig)

    def GetBindingDevice(self, sig=None):
        if self.get_value(bindings.BINDING) == bindings.DEVICE:
            self.on_update(BindingDevice, self.binding.addr.address, sig=sig)
        else:
            self.on_update(BindingDevice, None, sig=sig)

    def SetBindingDevice(self, address, sig=None, values=None):
        if address is None:
            address = self.binding.def_device

        try:
            address = int(address)
        except:
            raise dali.exceptions.CommandGenerateError('Value({}) must be integer!'.format(address))
        else:

            if self.get_value(BindingDevice) != address:
                # settings.dump.set_dump(self, BindingDevice, value)
                if self.get_value(bindings.BINDING) == bindings.DEVICE:
                    if values is None:
                        for _provider in self.manager.providers_by_address(address):
                            if isinstance(_provider, self.__class__):
                                values = _provider.values.copy()
                                break

                    if values is None:
                        self.reset()
                    else:
                        for funit_type, value in values.items():
                            self.on_update(funit_type, value)

            self.on_update(BindingDevice, address, sig=sig)

    def GetBindingGroup(self, sig=None):
        if self.get_value(bindings.BINDING) == bindings.GROUP:
            self.on_update(BindingGroup, self.binding.addr.group, sig=sig)
        else:
            self.on_update(BindingGroup, None, sig=sig)

    def SetBindingGroup(self, value, sig=None):
        try:
            value = int(value)
        except:
            raise dali.exceptions.CommandGenerateError('Value({}) must be integer!'.format(value))

        if self.get_value(BindingGroup) != value:
            if self.get_value(bindings.BINDING) == bindings.GROUP:
                self.binding.set_addr(value)
            self.on_update(BindingGroup, value, sig=sig)

    def GetGroups(self, sig=None):
        raise NotImplementedError()

    def SetGroups(self, val, sig=None):
        raise NotImplementedError()

    def Gtin(self, sig=None):
        self.on_update(Gtin, ''.join([hex(b)[2:].rjust(2, '0').upper() for b in self.read_memory_location(0, 0x3, 0x9)]), sig=sig)

    def FirmwareVersion(self, sig=None):
        self.on_update(FirmwareVersion, '.'.join([str(b) for b in self.read_memory_location(0, 0x9, 0xb)]), sig=sig)

    def Serial(self, sig=None):
        self.on_update(Serial, ''.join([hex(b)[2:].rjust(2, '0').upper() for b in self.read_memory_location(0, 0xb, 0x13)]), sig=sig)

    def HardwareVersion(self, sig=None):
        self.on_update(HardwareVersion, '.'.join([str(b) for b in self.read_memory_location(0, 0x13, 0x15)]), sig=sig)

    def GtinOem(self, sig=None):
        self.on_update(GtinOem, ''.join([hex(b)[2:].rjust(2, '0').upper() for b in self.read_memory_location(1, 0x3, 0x9)]), sig=sig)

    def SerialOem(self, sig=None):
        self.on_update(SerialOem, ''.join([hex(b)[2:].rjust(2, '0').upper() for b in self.read_memory_location(1, 0x9, 0x11)]), sig=sig)

    def serialize(self) -> dict:
        d = {}
        if self.is_device:
            d['address'] = self.binding.addr.address
            for funit_type in [Gtin, Serial, GtinOem, SerialOem, FirmwareVersion, HardwareVersion]:
                value = self.get_value(funit_type)
                key = str(funit_type)[0].lower() + str(funit_type)[1:]
                if value is not None:
                    d[key] = value
        return d


class DaliFunit(Funit):
    get_for_not_devices = True
    set_cmd = None
    get_cmd = None
    set_func = None
    get_func = None
    easy_get = False
    get_after_set = True
    is_control = False
    _dtr0 = None
    _dtr1 = None
    _dtr2 = None

    def _prepare_cmd(self, provider, cmd, *args):
        raise NotImplemented()

    @property
    def signature(self) -> CommandSignature:
        return CommandSignature(self.funit_type, session_id=self._variable.session_id, key=self._variable.key)

    def _apply_cmd(self, provider: DaliProvider, cmd, *args):
        cmds = []
        if cmd._uses_dtr1:
            cmds.append(self._dtr1(0))
            logging.warning('NEED DTR1 for %s of %s'.format(cmd, provider))

        if cmd._uses_dtr2:
            cmds.append(self._dtr2(0))
            logging.warning('NEED DTR2 for %s of %s'.format(cmd, provider))

        if cmd._uses_dtr0:
            cmds.append(self._dtr0(args[0]))
            args = args[1:]

        cmd_instance = self._prepare_cmd(provider, cmd, *args)

        cmd_instance.sig = self.signature
        cmds.append(cmd_instance)

        return cmds

    def _apply_func(self, entity, func):
        cmd = MultiCommand(getattr(entity, func), self.set_value, **self.params)
        cmd.set_signature(self.funit_type, session_id=self._variable.session_id, key=self._variable.key)
        return cmd

    def set(self):
        if self.set_cmd:
            cmds = self._apply_cmd(self.entity, self.set_cmd, self.set_value)
            self.entity.on_command_prepared(*cmds, is_control=self.is_control)
            if not self.entity.is_device:
                self.entity.on_update(self.funit_type, self.set_value)

        elif self.set_func:
            self.entity.manager.add_command(self.entity, self._apply_func(self.entity, self.set_func.__name__), is_control=self.is_control)

        if self.get_cmd and self.get_after_set:
            self.get()

    def get(self):
        if self.easy_get:
            if self.signed:
                self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
        else:
            if not self.entity.is_device:
                # return
                if self.get_for_not_devices:
                    self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
                    self._variable.clear_signature()
                else:
                    return

            for _p_id in self.entity.get_members():
                _provider = self.entity.manager.get_provider(_p_id)
                if isinstance(_provider, self.entity_type) and _provider.is_device:
                    if self.get_cmd:
                        _provider.on_command_prepared(*self._apply_cmd(_provider, self.get_cmd))
                    elif self.get_func:
                        _provider.manager.add_command(_provider, self._apply_func(_provider, self.get_func.__name__))


DaliProvider.funit_class = DaliFunit


class Address(DaliFunit):
    get_for_not_devices = False
    set_func = DaliProvider.SetShortAddress


class Discovery(DaliFunit):
    get_for_not_devices = False
    easy_get = True
    set_func = DaliProvider.SetDiscovery


class Groups(DaliFunit):
    set_func = DaliProvider.SetGroups
    get_func = DaliProvider.GetGroups


class Gtin(DaliFunit):
    get_for_not_devices = False
    get_func = DaliProvider.Gtin


class Serial(DaliFunit):
    get_for_not_devices = False
    get_func = DaliProvider.Serial


class GtinOem(DaliFunit):
    get_for_not_devices = False
    get_func = DaliProvider.GtinOem


class SerialOem(DaliFunit):
    get_for_not_devices = False
    get_func = DaliProvider.SerialOem


class FirmwareVersion(DaliFunit):
    get_for_not_devices = False
    get_func = DaliProvider.FirmwareVersion


class HardwareVersion(DaliFunit):
    get_for_not_devices = False
    get_func = DaliProvider.HardwareVersion


class Binding(DaliFunit):
    get_for_not_devices = False
    set_func = DaliProvider.SetBinding
    get_func = DaliProvider.GetBinding


class BindingDevice(DaliFunit):
    get_for_not_devices = False
    set_func = DaliProvider.SetBindingDevice
    get_func = DaliProvider.GetBindingDevice


class BindingGroup(DaliFunit):
    get_for_not_devices = False
    set_func = DaliProvider.SetBindingGroup
    get_func = DaliProvider.GetBindingGroup


class Validity(DaliFunit):
    easy_get = True
