from spread_core.errors.project_errors import ClassifierError, CommandError
from spread_core.mqtt.spread.topic import Set, Renew
from spread_core.mqtt.spread.variable import Variable
from spread_core.tools import settings
from spread_core.tools.utils import get_subclass


class Entity:
    funit_class = None

    def __init__(self, _id):
        self._id = _id
        self.values = dict()

    @property
    def id(self): return self._id

    def __str__(self): return '{}({})'.format(self.__class__.__name__, self.id)
    def __repr__(self): return self.__str__()

    def is_dumped(self, key) -> bool: return False
    def get_address(self, funit_type): raise NotImplementedError()
    def repair_on_fail(self) -> tuple: return ()
    def not_repaired(self) -> tuple: return self.repair_on_fail()

    def on_command(self, topic: (Set, Renew), variable: Variable):
        funit_type = topic.entity_address.funit_type
        command = self.funit_class.of(funit_type)
        if command and issubclass(command, Funit):
            try:
                if isinstance(topic, Set):
                    command(self, variable).set()
                elif isinstance(topic, Renew):
                    command(self, variable).get()
                else:
                    raise Funit.NotImplemented()
            except Funit.NotImplemented:
                raise ClassifierError.of_action(topic.__class__.__name__, self, funit_type)
            except BaseException as ex:
                raise CommandError.of_exec(str(ex))
        else:
            raise ClassifierError.of_funit(funit_type, self)

    def get_value(self, funit_type, def_value=None):
        if funit_type in self.values:
            return self.values[funit_type]
        else:
            _exist, _ow = settings.dump.get_dump(self.__class__.__name__, self.id, funit_type)
            if _exist:
                return _ow
            else:
                return def_value

    def set_value(self, key, value):
        self.values[key] = value
        if self.is_dumped(key):
            if isinstance(key, FunitMetaClass) or (isinstance(key, type) and issubclass(key, FunitMetaClass)):
                key = key.funit_type
            settings.dump.set_dump(self, key, value)


class FunitMetaClass(type):
    @property
    def funit_type(cls): return cls.__name__

    def __eq__(cls, other):
        if isinstance(other, FunitMetaClass):
            return other.funit_type == cls.funit_type

        return other == cls.funit_type

    def __str__(cls): return cls.funit_type
    def __hash__(cls): return hash(cls.funit_type)


class Funit(metaclass=FunitMetaClass):
    retained = True
    delay = 0
    entity_type = Entity
    variants = None
    index_of = []
    mask = None
    value_type = None
    value_range = None

    class NotImplemented(BaseException):
        pass

    def __init__(self, entity: entity_type, variable: Variable):
        self._entity = entity
        self._variable = variable
        self._sig = None
        self._set_value = None

    @classmethod
    def of(cls, funit_type): return get_subclass(cls, lambda cmd: issubclass(cmd, cls) and cmd.funit_type == funit_type)

    @property
    def funit_type(self): return self.__class__.funit_type

    @property
    def entity(self) -> entity_type: return self._entity

    @property
    def set_value(self):
        _value = self._variable.value
        if self.value_type:
            try:
                _value = self.value_type(_value)
            except:
                raise ValueError(f'В качестве value ожидается значение типа {self.value_type}')

        if self.index_of:
            if _value not in self.index_of:
                raise ValueError(f'В качестве value ожидается значение из {self.index_of}')
            _value = self.index_of.index(_value)
        elif self.variants and _value not in self.variants:
            raise ValueError(f'В качестве value ожидается значение из {self.variants}')
        elif self.mask:
            try:
                _value = sum([1 << self.mask.index(i) for i in set(_value)])
            except:
                raise ValueError(f'В качестве value ожидаются значения из {self.mask}')
        elif self.value_range and _value not in self.value_range:
            raise ValueError(f'В качестве value ожидаются значения из диапозона ['
                             f'{self.value_range.start}..{self.value_range.stop-1}]')

        return _value

    @property
    def params(self) -> dict: return self._variable.params

    @property
    def signed(self): return self._variable.session_id

    def set(self): raise self.NotImplemented()
    def get(self): raise self.NotImplemented()
    def state(self): raise self.NotImplemented()
