import datetime
import logging
from threading import Timer

from spread_core.tools.settings import config, DATA, FADE_TIME
from . import Cycle, Incubator, PeriodicalTrigger
from .handler import Handler
from ...mqtt.spread.variable import Variable


class Cycler(Handler):
    triggers_path = property(fget=lambda self: config[DATA])

    def __init__(self, project_id, _mqttc):
        super(Cycler, self).__init__(project_id, _mqttc)
        PeriodicalTrigger.FADE_TIME = config[FADE_TIME]
        self.timer = None
        self.triggers = []

    def on_ready(self):
        self.on_timer()

    def parse(self, data):
        cycles = {
            cycle_data['id']: Cycle.of(cycle_data)
            for cycle_data in data['cycles']
        }

        [
            Incubator.of(loc_data, cycles[loc_data['cycle_id']])
            for loc_data in data['locations'] if loc_data['cycle_id'] in cycles
        ]

        for trigger_data in data['triggers']:
            cycle = cycles.get(trigger_data['cycle_id'], None)
            if isinstance(cycle, Cycle) and isinstance(cycle.incubator, Incubator) and cycle.incubator.active:
                if trigger_data['topic'] and trigger_data['day'] < cycle.period:
                    trigger = PeriodicalTrigger.of(trigger_data, cycle)
                    cycle.set_trigger(trigger)

        for cycle in cycles.values():
            if cycle.incubator:
                cycle.sort()
                last_trigger = None
                for trigger in cycle.triggers:
                    if isinstance(trigger, PeriodicalTrigger):
                        if last_trigger:
                            trigger.level_from = last_trigger.level
                            last_trigger.next_time = trigger.datetime
                        else:
                            trigger.level_from = cycle.startLevel

                        last_trigger = trigger
                        if trigger.has_next:
                            self.triggers.append(trigger)

    def on_timer(self):
        try:
            self.triggers.sort()
            if len(self.triggers) > 0:
                trigger = self.triggers[0]
                interval = trigger.time_left + 1
                logging.info(f'SET TIMER to {datetime.timedelta(seconds=interval)} sec FOR {trigger}')
                if self.timer:
                    self.timer.cancel()
                self.timer = Timer(interval=interval, function=self.exec_task, args=[trigger])
                self.timer.start()
            else:
                logging.info('COMPLETE')
                self.on_exit()
        except BaseException as ex:
            logging.exception(ex)

    def exec_task(self, _: PeriodicalTrigger):
        for _trigger in self.triggers:
            if _trigger.time_left > 0: break
            logging.info(f'EXECUTE {_trigger}:')
            progress = (datetime.datetime.now() - _trigger.datetime).total_seconds() / _trigger.cycle.duration / 60
            progress = min(1, max(0, progress))
            value = int(_trigger.level_from + (_trigger.level - _trigger.level_from)*progress)
            self.publish(_trigger.topic, Variable(value), log_level=logging.INFO)
            _trigger.on_execute(progress)
            if not _trigger.has_next:
                self.triggers.remove(_trigger)

        self.on_timer()

    def on_exit(self):
        super(Cycler, self).on_exit()
        if self.timer:
            self.timer.cancel()
            self.timer = None
