import datetime
from .triggers import DateTrigger


class Cycle:
    __slots__ = ['id', 'name', 'period', 'duration', 'startLevel', '_incubator', '_triggers']

    def __init__(self):
        self._incubator = None
        self._triggers = []

    @property
    def incubator(self): return self._incubator

    @incubator.setter
    def incubator(self, incubator): self._incubator = incubator

    def set_trigger(self, trigger):
        self._triggers.append(trigger)

    @property
    def triggers(self): return self._triggers

    def sort(self):
        self._triggers.sort(key=lambda tr: tr.datetime)

    @classmethod
    def of(cls, data: dict):
        self = cls()
        for attname, value in data.items():
            if attname in self.__slots__:
                setattr(self, attname, value)
        return self


class Incubator:
    __slots__ = ["id", "location_id", "startTime", "_cycle", "active", "recursive"]

    def __init__(self):
        self._cycle = None

    @property
    def cycle(self) -> Cycle: return self._cycle

    @cycle.setter
    def cycle(self, cycle: Cycle):
        cycle.incubator = self
        self._cycle = cycle

    @classmethod
    def of(cls, data: dict, cycle: Cycle):
        self = cls()
        self.cycle = cycle
        for attname, value in data.items():
            if attname in self.__slots__:
                setattr(self, attname, value)
        return self


class PeriodicalTrigger(DateTrigger):
    FADE_TIME = 2

    def __init__(self, data):
        super(PeriodicalTrigger, self).__init__(**data)
        self.day = None
        self.time = None
        self._cycle = None
        self._datetime = None
        self._action_datetime = None
        self.topic = None
        self.actions = []
        self.topic = None
        self.level_from = None
        self.level = None
        self.next_time = None

    def __str__(self): return '<[%s]%s>(%s день %s)%s..%s' % (self.cycle.incubator.location_id, self.cycle.name, self.day, self.datetime.time(), self.level_from, self.level)

    @property
    def has_next(self): return self._action_datetime is not None or self._datetime > datetime.datetime.now() or self.cycle.incubator.recursive

    @property
    def time_left(self):
        if self._action_datetime:
            return int((self._action_datetime - datetime.datetime.now()).total_seconds())

        return int((self.datetime - datetime.datetime.now()).total_seconds())

    def on_execute(self, progress=0):
        next_action = datetime.datetime.now() + datetime.timedelta(seconds=self.FADE_TIME)
        if progress < 1 and (self.next_time is None or next_action < self.next_time):
            self._action_datetime = next_action
        else:
            self._action_datetime = None

    @property
    def incubator(self) -> Incubator: return self.cycle.incubator

    @property
    def cycle(self) -> Cycle: return self._cycle

    @cycle.setter
    def cycle(self, cycle: Cycle): self._cycle = cycle

    @property
    def datetime(self):
        if self._datetime is None:
            self._datetime = datetime.datetime.combine(
                datetime.date.fromisoformat(self.incubator.startTime),
                datetime.time.fromisoformat(self.time)
            ) + datetime.timedelta(days=self.day)

        now = datetime.datetime.now()
        if self._action_datetime is None and now > self._datetime and self.cycle.incubator.recursive:
            self._datetime += datetime.timedelta(days=1 + self.cycle.period * (now - self._datetime).days)
        return self._datetime

    @classmethod
    def of(cls, data: dict, cycle: Cycle):
        self = cls(data)
        data.pop('id')
        for attname, value in data.items():
            if hasattr(self, attname):
                setattr(self, attname, value)

        self.cycle = cycle

        return self
