import json
import logging
import threading

import spread_core.mqtt as mqtt
from spread_core.bam import generator as generator
from spread_core.bam.entities import Entity
from spread_core.bam.multi_command import MultiCommand
from spread_core.errors.project_errors import ProjectError, JocketError, InvalidDevice, InitError, ClassifierError, \
    CommandError
from spread_core.protocols.dali.command import Command
from spread_core.protocols.dali.exceptions import CommandGenerateError
from spread_core.tools.manager_interface import ManagerOfBroker
from spread_lom.commands import LCommand
from spread_mercury.commands import MCommand

CONTROL = 4
COMMAND = 3
TUNER = 2
INFO = 1
SURVEY = 0


class Manager(Entity, ManagerOfBroker):
    STATUS_STOPPED = 'Stopped'
    STATUS_RUNNING = 'Running'
    STATUS_INFO = 'Info'
    STATUS_SURVEY = 'Survey'

    send_interface = None
    _server_id = None
    _key = None
    _bus_id = None
    _pollRate = None
    _survey_timer = None
    _survey_event = threading.Event()
    _step_event = threading.Event()
    _event = threading.Event()
    _external_cmd = threading.Event()
    INTERRUPT_SURVEY = True

    server_id = property(fget=lambda self: self._server_id)

    def __init__(self, p_id,  m_id):
        Entity.__init__(self, m_id)
        self._stopped = False
        self._providers = dict()
        self._addresses = dict()
        self._subgineries = dict()
        self._engineries = dict()
        self.project_id = p_id
        self.actions = {CONTROL: [], COMMAND: [], TUNER: [], INFO: [], SURVEY: []}
        self.ready_timer = None
        self.bus = None
        self.name = ''

    def _sorted_info(self):
        return self._providers

    @property
    def providers(self):
        """returns copy of providers"""
        return self._providers.copy()

    def get_provider(self, provider_id):
        """returns provider by id"""
        return self._providers.get(provider_id, None)

    def providers_by_address(self, short_address) -> list:
        """returns provider by address"""
        provider_ids = self._addresses.get(short_address, [])
        return [self._providers.get(provider_id, None) for provider_id in provider_ids]

    @property
    def addresses(self):
        """returns copy of addresses"""
        return self._addresses.copy()

    def set_addresses(self, address: int, participants: list):
        """set participants of address"""
        if isinstance(participants, (list, tuple)):
            self._addresses[address] = list(participants)

    def parse(self, data):
        if data['type'] != self.__class__.__name__:
            raise InitError(f'Manager type required as "{self.__class__.__name__}" but set "{data["type"]}"')
        self._server_id = data['serverID']
        self.bus = data['attributes'].get('bus', None)
        self.name = data.get('name', f'id:{self.id}')
        self._pollRate = data['attributes']['pollRate']
        self.SetStatus(None, self.STATUS_RUNNING)

    def set_broker(self, mqttc, send_interface):
        ManagerOfBroker.__init__(self, mqttc)
        send_interface.set_event_handler(self._event)
        send_interface.set_external_cmd_handler(self._external_cmd)
        self.send_interface = send_interface

    def start(self):
        self.subscribe(mqtt.project.TopicProject.managers(self.project_id))

    def on_state(self, topic: mqtt.spread.topic.SpreadTopic = None, variable: mqtt.spread.variable.Variable = None):
        if topic and variable:
            if self.ready_timer:
                self.ready_timer.cancel()
            if isinstance(topic.entity_address, mqtt.spread.address.ProviderAddress):
                _p_id = topic.entity_address.id
                if _p_id in self._providers:
                    if variable.value is not None:
                        if topic.entity_address.funit_type not in self._providers[_p_id].not_repaired():
                            self._providers[_p_id]._invalid_values[topic.entity_address.funit_type] = variable.value

        self.ready_timer = threading.Timer(1, self.on_ready)
        self.ready_timer.setName('on_ready Timer')
        self.ready_timer.start()

    def on_project(self, topic, data):
        self.unsubscribe(topic)
        payload = json.loads(data)
        if mqtt.project.TopicProject.MANAGERS in payload:
            for item in payload[mqtt.project.TopicProject.MANAGERS]:
                if item['id'] == self.id:
                    try:
                        self.parse(item)
                    except BaseException as ex:
                        raise InitError(str(ex))
                    self.subscribe(mqtt.project.TopicProject.providers(self.project_id), True)
                    return
            raise ProjectError("В файле 'managers.json' отсутствует менеджер с ID={}".format(self.id))
        elif mqtt.project.TopicProject.PROVIDERS in payload:
            for item in payload[mqtt.project.TopicProject.PROVIDERS]:
                if item['managerID'] == self.id:
                    self.add_provider(item, self)
            logging.info('Project({}) loaded'.format(self.project_id))
            self.on_project_ready()

    def on_project_ready(self):
        self.on_state()
        self.subscribe(mqtt.spread.topic.State(self.get_address('+/+/#')))

    def add_provider(self, data, manager):
        provider = generator.generate_provider(data)
        provider.set_manager(manager)
        self._providers[provider.id] = provider

    def on_ready(self):
        self.unsubscribe(mqtt.spread.topic.State(self.get_address('+/+/#')))
        self.subscribe(mqtt.spread.topic.Set(self.get_address(None)))
        self.subscribe(mqtt.spread.topic.Renew(self.get_address(None)))

        if self.send_interface:
            self.send_interface.start()

        threading.Thread(target=self.info, name='InfoThread').start()

    def on_exit(self):
        super(Manager, self).on_exit()
        self.send_interface.set_stopped()

        self._stopped = True
        self.actions = {CONTROL: [], COMMAND: [], TUNER: [], INFO: [], SURVEY: []}

        if self.mqttc.connected:
            for p_id in self._providers:
                provider = self._providers[p_id]
                provider.on_exit()

            self.SetStatus(None, self.STATUS_STOPPED)

        if self._survey_timer:
            self._survey_timer.cancel()
        self._survey_event.set()
        self._step_event.set()
        self._event.set()
        self._external_cmd.set()

    def survey(self):
        while self._stopped is False:
            self._survey_event.wait()
            if len(self.actions[SURVEY]) == 0:
                for p_id in self._providers:
                    provider = self._providers[p_id]
                    if provider.surveyable:
                        provider.on_survey()
            if not self._step_event.is_set():
                self._step_event.set()
            self._survey_event.clear()
            if self._stopped is False:
                self._survey_timer = threading.Timer(interval=self._pollRate / 1000, function=self.dispatch_survey)
                self._survey_timer.setName('dispatch_survey Timer')
                self._survey_timer.start()

    def dispatch_survey(self):
        if not self._survey_event.is_set():
            self._survey_event.set()

    def external_command_handler(self):
        while self._stopped is False:
            command_frame = self.send_interface.get_ex_cmd()
            if command_frame:
                try:
                    self.on_external_command(command_frame)
                except BaseException as ex:
                    logging.debug(f'EXTERNAL FRAME ERROR[{command_frame}]')
                    logging.exception(ex)
            else:
                self._external_cmd.clear()
                self._external_cmd.wait()

    def on_external_command(self, command_frame):
        raise NotImplementedError()

    def event_handler(self):
        while not self._stopped:
            event = self.send_interface.get_event()
            if event:
                try:
                    self.on_event(event)
                except BaseException as ex:
                    logging.debug(f'EVENT FRAME ERROR[{event}]')
                    logging.exception(ex)

            else:
                self._event.clear()
                self._event.wait()

    def on_event(self, event):
        raise NotImplementedError()

    def info(self):
        self.SetStatus(None, self.STATUS_INFO)
        for p_id in self._sorted_info():
            if self._stopped:
                return
            provider = self._providers[p_id]
            if provider.infoble:
                try:
                    provider.get_bindings()
                    provider.check_valid()
                    provider.commit_states(provider.is_valid)
                    if provider.is_valid:
                        provider.get_info()
                        provider.override_info()
                except InvalidDevice as ex:
                    logging.warning(ex)
                    provider.set_valid(provider.INVALID)

        threading.Thread(target=self.survey, name='SurveyThread').start()
        threading.Thread(target=self.do_next, name='DoNextThread').start()
        threading.Thread(target=self.external_command_handler, name='ExternalThread').start()
        threading.Thread(target=self.event_handler, name='EventThread').start()
        self._survey_event.set()
        self.SetStatus(None, self.STATUS_SURVEY)

    def do_next(self):
        while self._stopped is False:
            self._step_event.wait()

            _priority = -1
            stack = ()
            for _priority, stack in self.actions.items():
                if stack: break

            while len(stack) > 0:
                need_break = False
                for _p, s in self.actions.items():
                    if _p > _priority and len(s) > 0:
                        need_break = True
                        break

                if need_break: break

                provider, cmd = stack.pop(0)

                if isinstance(cmd, MultiCommand):
                    try:
                        cmd.call()
                    except BaseException as ex:
                        logging.exception(ex)
                elif isinstance(cmd, (Command, MCommand)):
                    try:
                        if provider:
                            provider.request(cmd)
                        else:
                            self.send_interface.send(cmd)
                    except BaseException as ex:
                        logging.exception(ex)
                elif isinstance(cmd, LCommand):
                    try:
                        if provider:
                            provider.request(cmd, mc=not provider.is_device)
                        else:
                            self.send_interface.send(cmd, mc=not provider.is_device)
                    except BaseException as ex:
                        logging.exception(ex)
                else:
                    logging.warning('unknown command: {}'.format(cmd))

                if _priority == SURVEY and self.INTERRUPT_SURVEY:
                    break

            self._step_event.clear()

    def add_command(self, provider, *cmds, is_control=False, is_tuner=False):
        if not self._stopped:
            for cmd in cmds:
                if cmd:
                    if is_control:
                        self.actions[CONTROL].append((provider, cmd))
                    elif is_tuner:
                        self.actions[TUNER].append((provider, cmd))
                    else:
                        self.actions[COMMAND].append((provider, cmd))
                else:
                    logging.warning('Empty command')

            if not self._step_event.is_set():
                self._step_event.set()

    def add_info(self, provider, cmd):
        if self._stopped:
            return
        if (provider, cmd) not in self.actions[INFO]:
            self.actions[INFO].append((provider, cmd))

    def add_survey(self, provider, cmd):
        if self._stopped:
            return
        if (provider, cmd) not in self.actions[SURVEY]:
            self.actions[SURVEY].append((provider, cmd))

    def on_command(self, topic: mqtt.spread.topic.Set, variable: mqtt.spread.variable.Variable):
        self.log_command(topic, variable)
        funit_type = topic.entity_address.funit_type
        if self._stopped:
            logging.warning('Manager on scan/extend/construct. Reject command!')
            return
        try:
            act = 'set' if isinstance(topic, mqtt.spread.topic.Set) else 'get'
            if topic.entity_address.id == self.id:
                funit = self.get_funit(funit_type)
                if act in funit:
                    if 'func' in funit[act] and funit[act]['func'] in dir(self):
                        cmd = MultiCommand(getattr(self, funit[act]['func']), variable.params)
                        cmd.set_signature(funit_type, variable.key, variable.session_id)
                        self.add_command(self, cmd)
                    else:
                        raise JocketError(
                            '<<{}>> is not implemented in classifier for <<{}>>!'.format(act, self.__class__.__name__))
                else:
                    raise JocketError(
                        '<<{}>>  of <<{}>> is not present in <<{}>>!'.format(act, funit_type, self.__class__.__name__))
            elif topic.entity_address.id in self._providers:
                provider = self._providers[topic.entity_address.id]
                if not provider.is_valid and act != mqtt.const.GET and 'Binding' not in funit_type:
                    raise CommandGenerateError('device is invalid!')
                provider.on_command(topic, variable)
            else:
                raise JocketError('объект с адресом {} не найден!'.format(topic.entity_address.id))
        except (JocketError, CommandGenerateError, ClassifierError, InvalidDevice, CommandError) as ex:
            logging.warning(ex)
            if variable.session_id:
                to_topic = mqtt.spread.topic.Reply(self.project_id, variable.session_id)
                self.publish(to_topic, str(ex))

        if not self._step_event.is_set():
            self._step_event.set()

    def log_command(self, topic: mqtt.spread.topic.SpreadTopic, variable: mqtt.spread.variable.Variable):
        logging.debug(
            f'COMMAND: {topic.__class__.__name__} {topic.entity_address.type}({topic.entity_address.id})'
            f'.{topic.entity_address.funit_type}' +
            (f'={variable.value}' if isinstance(topic, mqtt.spread.topic.Set) else '') +
            (f' params: {variable.params}' if variable.params else '') +
            f' queue: {"+".join([str(len(q)) for q in self.actions.values()])}'
        )

    def SetStatus(self, value, sig=None):
        if self.send_interface:
            self.on_update('Status', value)

    def get_funit(self, funit_type):
        if self.__class__.__name__ in mqtt.classifier:
            if funit_type in mqtt.classifier[self.__class__.__name__]:
                return mqtt.classifier[self.__class__.__name__][funit_type]
        return None

    def get_address(self, funit_type):
        if funit_type is None: funit_type = '#'
        return mqtt.spread.address.ManagerAddress.of(self, funit_type)

    def on_update(self, funit_type, value, retain=True):
        _topic = mqtt.spread.topic.State(self.get_address(funit_type))
        _variable = mqtt.spread.variable.Variable(value)
        self.publish(topic=_topic, data=_variable, retain=retain)
