import asyncio
import json
import logging
import os
import shutil
import zipfile
import zlib

import aiofiles
import aiomqtt

from spread_core.bam.engineries import lighting_area as la
from spread_core.protocols import dali
from spread_core import mqtt
from spread_core.tools import settings

CONNECT_RESULTS = [
    'Connection successful',
    'Connection refused - incorrect protocol version',
    'Connection refused - invalid client identifier',
    'Connection refused - server unavailable',
    'Connection refused - bad username or password',
    'Connection refused - not authorised',
]


def check_la_btn_preset_action(data):
    if data['event'] not in dali.device.events.ButtonEvent.STATES.values():
        raise ValueError('event ожидается как один из %s' % dali.device.events.ButtonEvent.STATES.values())

    if data['command'] not in set(la.VacancyAction.variants
                                  + la.OccupancyAction.variants
                                  + [
                                      la.ACTION_UP,
                                      la.ACTION_DOWN,
                                      la.ACTION_STEP_UP,
                                      la.ACTION_STEP_DOWN,
                                      la.ACTION_ON_AND_STEP_UP,
                                      la.ACTION_STEP_DOWN_AND_OFF,
                                      la.EMULATE_OCCUPANCY,
                                      la.EMULATE_VACANCY,
                                  ]):
        raise ValueError('указано некорректное значение для "command"')

    if data['command'] == la.ACTION_SCENE and (data['scene'] not in range(0, 16) and data['scene'] != 255):
        raise ValueError('scene ожидается как 0..15 или 255')

    if data['command'] == la.ACTION_LEVEL and data['level'] not in range(0, 256):
        raise ValueError('level ожидается как 0..255')

    if data.get('setLightingArea', None) not in ['On', 'Off', None]:
        raise ValueError('"setLightingArea" не ожидается или ожидается как "On"/"Off"')

    if not isinstance(data.get('stopTuning', False), bool):
        raise ValueError('"stopTuning" ожидается как bool')


class AuthError(BaseException):
    @staticmethod
    def connection_result(rc):
        try: return CONNECT_RESULTS[rc]
        except IndexError: return 'Connect result is %s' % rc

    def __init__(self, rc):
        super(AuthError, self).__init__(self.connection_result(rc))


class Projector:
    def __init__(self, loop=None):
        self._loop = loop
        self._client = None
        self.project_hash = None
        self.connected = asyncio.Event(loop=self.loop)
        self.hash_checked = False
        self.local_checked = False

        self._host = settings.config[settings.BROKER][settings.HOST]
        self._port = settings.config[settings.BROKER][settings.PORT]
        self._login = settings.config[settings.BROKER][settings.LOGIN]
        self._password = settings.config[settings.BROKER][settings.PASSWORD]

    @property
    def project_id(self): return settings.config[settings.PROJECT_ID]

    @property
    def loop(self):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()
        return self._loop

    @property
    def client(self) -> aiomqtt.Client:
        if self._client is None:
            c = aiomqtt.Client(self.loop, userdata=self.__class__.__name__)
            c.username_pw_set(username=self._login, password=self._password)
            c.loop_start()  # See "About that loop..." below.

            c.on_connect = self.on_connect
            c.on_message = self.on_message
            c.on_disconnect = self.on_disconnect
            self.connected.clear()
            self._client = c
        return self._client

    async def manual_check_hash(self):
        await asyncio.sleep(10)
        if not self.hash_checked:
            await self.check_hash(-1)

    async def start(self):
        await self.client.connect(host=self._host, port=self._port)
        await self.connected.wait()
        self.loop.create_task(self.manual_check_hash())

    def disconnect(self):
        if self._client:
            self._client.on_disconnect = None
            self._client.disconnect()
            self.loop.create_task(self._client.loop_stop())
            self._client = None

    def on_connect(self, client, userdata, flags, rc):
        if rc == 0:
            self.connected.set()
            self.local_checked = False
            self.subscribe(mqtt.project.LAButtonPresets.broadcast(self.project_id))
            self.subscribe(mqtt.project.ProjectHash(self.project_id))
            self.loop.create_task(self.manual_check_local())
        else:
            raise AuthError(rc)

    def on_disconnect(self, client, userdata, rc):
        raise AuthError(rc)

    def subscribe(self, topic: (str, mqtt.project.TopicProject)):
        self.client.subscribe(str(topic))
        logging.debug('subscribed to %s' % topic)

    def unsubscribe(self, topic: (str, mqtt.project.TopicProject)):
        self.client.unsubscribe(str(topic))
        logging.debug('unsubscribed to %s' % topic)

    async def publish_async(self, topic: (mqtt.Topic, str), data, retain=False, logged=True):
        await self.publish(topic, data, retain=retain, logged=logged).wait_for_publish()

    def publish(self, topic: (str, mqtt.Topic), data, retain=False, logged=True):
        if not isinstance(data, (int, str, bytes, bytearray)):
            data = json.dumps(data)

        if logged:
            if len(str(data)) > 1024:
                logging.debug('[%s] -> %s' % (topic, '%s bytes' % len(str(data))))
            else:
                logging.debug('[%s] -> %s' % (topic, data))
        return self.client.publish(str(topic), data, retain=retain)

    def on_message(self, client, userdata, message):
        topic = mqtt.of(message.topic)
        data = message.payload.decode()
        try:
            data = json.loads(data)
        except json.JSONDecodeError:
            pass
        except BaseException as ex:
            logging.exception(ex)

        if isinstance(topic, mqtt.project.ProjectHash):
            self.hash_checked = True
            self.loop.create_task(self.check_hash(data))
        elif isinstance(topic, mqtt.project.LAButtonPresets):
            self.local_checked = True
            self.loop.create_task(self.check_local(topic, data))

        logging.debug('[%s] <- %s' % (message.topic, data))

    async def check_hash(self, current_hash):
        zip_path = settings.config.get(settings.CTP_PATH, None)
        if zip_path is None:
            raise BaseException(f'"{settings.CTP_PATH}" is not set!')

        async with aiofiles.open(zip_path, mode='rb') as zip_file:
            zip_data = await zip_file.read()

        real_hash = zlib.crc32(zip_data)

        if current_hash != real_hash:
            self.disconnect()
            logging.info('mosquitto clean...')

            os.system('sudo service mosquitto stop')
            db = settings.config.get(settings.MOSQUITTO_PATH, '/var/lib/mosquitto/')
            if os.path.isdir(db):
                shutil.rmtree(db, ignore_errors=True)
            elif os.path.isfile(db):
                os.remove(db)
            os.system('sudo service mosquitto start')
            logging.info('mosquitto cleaned')

            await self.client.connect(host=self._host, port=self._port)
            await self.connected.wait()

            logging.info('run filling dashboard')
            os.system('sudo spread_utils_dashboard fill --all')
            logging.info('filling OK')

            input_zip = zipfile.ZipFile(zip_path, 'r')
            project_id = json.loads(input_zip.read('header.json'))['header']['project']['id']

            for file in input_zip.namelist():
                if file != mqtt.project.TopicProject.MODELS + '.json':
                    await self.publish_async(
                        topic=mqtt.project.TopicProject(project_id, file),
                        data=json.loads(input_zip.read(file)),
                        retain=True
                    )

            await self.publish_async(mqtt.project.TopicProject.ctp(project_id), zip_data, retain=True)
            await self.publish_async(mqtt.project.ProjectHash(project_id), real_hash, retain=True)
            logging.info('Project updated')
        else:
            logging.info('Projects is the same')

    async def check_local(self, topic, data):
        file_path = os.path.join(
            settings.config.get(settings.LOCAL_PATH),
            str(topic).split('%s/' % self.project_id)[1]
        ) + os.extsep + 'json'
        try:
            async with aiofiles.open(file_path, mode='r') as fs:
                preset = json.loads(await fs.read())
        except FileNotFoundError:
            preset = {}

        if preset != data:
            if data == '':
                if os.path.exists(file_path):
                    os.remove(file_path)
            else:
                try:
                    if data['name'] == '':
                        raise ValueError('"name" не может быть пустым')

                    if not isinstance(data.get('isToggle', False), bool):
                        raise ValueError('"isToggle" Ожидается как bool')

                    for handler in data['handlers']:
                        for action in handler:
                            check_la_btn_preset_action(action)

                except BaseException as ex:
                    logging.debug(ex)
                    if data.get('name'):
                        preset.update({
                            key: data.get(key, default)
                            for key, default in (
                                ('name', 'AutoName'),
                                ('isToggle', False),
                                ('handlers', []),
                            )})

                        for handler in preset['handlers'].copy():
                            for action in handler.copy():
                                try:
                                    check_la_btn_preset_action(action)
                                except:
                                    handler.remove(action)

                        self.publish(topic, preset, retain=True)
                else:
                    if not os.path.exists(os.path.dirname(file_path)):
                        os.makedirs(os.path.dirname(file_path), exist_ok=True)

                    async with aiofiles.open(file_path, mode='w') as fs:
                        await fs.write(json.dumps(data, indent=2))

    async def manual_check_local(self):
        await asyncio.sleep(10)
        if self.local_checked:
            return
        root = os.path.join(settings.config.get(settings.LOCAL_PATH), mqtt.project.Local.protocol)
        for dirpath, dirnames, filenames in os.walk(root):
            for file in filenames:
                try:
                    async with aiofiles.open(os.path.join(dirpath, file), mode='r') as fs:
                        preset = json.loads(await fs.read())
                    self.publish(
                        mqtt.project.LAButtonPresets(self.project_id, int(file.split(os.path.extsep)[0])),
                        preset,
                        retain=True
                    )
                except BaseException as ex:
                    logging.error(ex)


def run():
    loop = asyncio.get_event_loop()
    projector = Projector(loop)
    try:
        loop.run_until_complete(projector.start())
        loop.run_forever()
    except (AuthError, ConnectionRefusedError) as ex:
        logging.error(ex)
    except KeyboardInterrupt:
        logging.info('%s breaked' % projector.__class__.__name__)
    finally:
        loop.stop()


if __name__ == '__main__':
    run()
