import logging
import threading

import can

from spread_core.errors.project_errors import InitError
from spread_core import mqtt
from .. import frame
from ..command import Command, YesNoResponse
from ..frame import BackwardModuleError
from ..gear import general as gear

_can_logger = logging.getLogger('can.interfaces.socketcan.socketcan')
_can_logger.level = logging.WARNING
_can_logger.getChild("rx").level = logging.WARNING

DOUBLE_SEND_FLAG = 'D'
THREE_BYTES_FLAG = 'F'
FORCE_ANSWER_FLAG = 'W'
DALI_ERROR_FLAG = 'E'


class DaliInterface:
    _is_busy = 0
    _response = None
    _request_event = threading.Event()
    _last_command = None
    _events = []
    _external = []
    controller_id = None
    module_id = None
    channel = None
    event_handler = None
    external_cmd_handler = None

    def __init__(self, timeout=1):
        self.timeout = timeout
        self._stopped = False

    def set_stopped(self):
        self._stopped = True

    def set_bus_params(self, send_topic):
        raise NotImplemented()

    def send(self, command: Command):
        raise NotImplemented()

    def on_dump(self, response, *args, **kwargs):
        command = self._last_command
        if command:
            if response == command.frame.pack:
                command.on_commit()
                logging.debug('  -> ✓: {}'.format(str_bytes(response)))
                if command.is_commited and not command.is_query:
                    self.release_condition()
            elif len(response) == 3:
                if response[0] & 1 == 0:
                    self.on_dali2_event(response)
                else:
                    self.on_unknown_pack(response)
            elif not command.is_commited:
                self.on_unknown_pack(response)
            else:
                command.on_response()
                try:
                    if 'E' in kwargs.get('flags', []):
                        result = command.response(frame.BackwardModuleError())
                    elif len(response) == 0:
                        if command.response == YesNoResponse:
                            result = command.response(None)
                        else:
                            result = command.response(frame.BackwardModuleError())
                    else:
                        result = command.response(frame.BackwardFrame(response))

                    self._response = result

                    _flags_str = 'flags: [{}]'.format(', '.join(kwargs['flags'])) if kwargs.get('flags', []) else ''
                    logging.debug('    <- {}'.format('; '.join([str(result), _flags_str])))
                except BaseException as ex:
                    logging.error('[{}]: {}'.format(response, str(ex)))
                finally:
                    self.release_condition()
        elif len(response) == 3 and response[0] & 1 == 0:
            self.on_dali2_event(response)
        elif len(response) > 0:
            self.on_unknown_pack(response)

    def on_unknown_pack(self, pack):
        if self.external_cmd_handler:
            self._external.append(pack)
            if not self.external_cmd_handler.is_set():
                self.external_cmd_handler.set()
            return

        logging.debug('UNKNOWN FRAME {}'.format(str_bytes(pack)))

    def on_dali2_event(self, response, flags=()):
        if self.event_handler:
            self._events.append(response)
            if not self.event_handler.is_set():
                self.event_handler.set()

    def release_condition(self):
        self._request_event.set()

    def set_event_handler(self, event_handler):
        self.event_handler = event_handler

    def set_external_cmd_handler(self, external_cmd_handler):
        self.external_cmd_handler = external_cmd_handler

    def get_event(self):
        if len(self._events) > 0:
            return self._events.pop(0)
        return None

    def get_ex_cmd(self) -> bytes:
        if len(self._external) > 0:
            return self._external.pop(0)
        return b''


class BrokerAdapter(DaliInterface):
    def __init__(self, mqtt_client, timeout=1):
        super(BrokerAdapter, self).__init__(timeout)
        self._mqtt_client = mqtt_client
        self.send_topic = None
        self.dump_topic = None

    def start(self):
        if self.dump_topic:
            self._mqtt_client.subscribe(str(self.dump_topic))
        else:
            raise InitError('SendTopic not set!')

    def set_bus_params(self, bus):
        self.send_topic = str(mqtt.bus.TopicBus.of_send(bus))
        self.dump_topic = str(mqtt.bus.TopicBus.of_dump(bus))

    def on_dump(self, response, flags='', *args, **kwargs):
        flags = flags.split(':') if len(flags) > 0 else []
        response = bytes.fromhex(response) if len(response) > 0 else b''

        super(BrokerAdapter, self).on_dump(response, flags=flags)

    def send(self, command: Command):
        if self._stopped is True:
            return BackwardModuleError()

        assert isinstance(command, Command)

        if isinstance(command, gear._GearCommand) and command.device_type > 0:
            self.send(gear.EnableDeviceType(command.device_type))

        message = command.frame.pack

        data = ''.join(hex(b).replace('0x', '').rjust(2, '0') for b in message).upper()

        flags = []
        if command.is_config:
            flags += DOUBLE_SEND_FLAG
        if command.is_query:
            flags += FORCE_ANSWER_FLAG

        if len(flags) > 0:
            data += '#' + ':'.join(flags)

        command.on_send()
        self._last_command = command
        self._response = None

        self._mqtt_client.publish(self.send_topic, data)
        logging.debug(u"-> {}".format(command))

        self._request_event.wait(self.timeout)
        if self._last_command and self._last_command.is_query and self._response is None:
            self._response = command.response(frame.BackwardModuleError())
            logging.warning(f'      <- NOT ANSWERED {self._last_command}')
        self._last_command = None
        self._request_event.clear()

        if (isinstance(self._response, BackwardModuleError) or self._response is None or self._response._value is None) and command.need_retry:
            logging.debug(f'RESEND {command}')
            self._response = self.send(command)

        return self._response


class DaliCanAdapter(DaliInterface, threading.Thread):
    cmd = False
    cl = 1
    CONTROLLER_ID = 31

    def __init__(self, *args, **kwargs):
        super(DaliCanAdapter, self).__init__(*args, **kwargs)
        threading.Thread.__init__(self)
        self.name = self.__class__.__name__
        self._bus = None
        self.can_name = None
        self.module = None
        self.channel = None
        self.listener = None

    def set_stopped(self):
        super(DaliCanAdapter, self).set_stopped()
        self.bus.shutdown()

    def set_bus_params(self, send_topic):
        self.can_name, self.module, self.channel = (int(i) for i in send_topic.split('/') if i.isdigit())
        self.can_name = ('can%s' if self.can_name == 0 else 'rcan%s') % self.can_name

    @property
    def bus(self) -> can.Bus:
        if self.module is None or self.channel is None or self.can_name is None:
            raise NotImplementedError()
        if self._bus is None:
            self._bus = can.Bus(
                channel=self.can_name,
                **dict(
                    bustype='socketcan',
                    can_filters=[
                        {
                            "can_id": (self.module << 5) + 31,
                            "can_mask": 1023,
                        },
                    ]
                )
            )
        return self._bus

    def run(self):
        while self._stopped is False:
            try:
                logging.info('Start listening %s' % self.can_name)
                for msg in self.bus:
                    try:
                        if msg.data[0] == 0x01 and (msg.data[1] & 0x07) + 1 == self.channel:
                            self.on_dump(msg.data[2:], flags='E' if (msg.data[1] >> 5) & 1 == 1 else '')
                    except BaseException as ex:
                        logging.exception(ex)
            except (BaseException, KeyboardInterrupt) as ex:
                logging.info('Stop listening %s. %s' % (self.can_name, str(ex)))
                if isinstance(ex, KeyboardInterrupt):
                    self._stopped = True

    def send(self, command: Command):
        if isinstance(command, gear._GearCommand) and command.device_type > 0:
            self.send(gear.EnableDeviceType(command.device_type))

        command.on_send()
        self._last_command = command
        self._response = None

        self.bus.send(can.Message(
            arbitration_id=(DaliCanAdapter.CONTROLLER_ID << 5) + self.module,
            data=bytes((
                (1 << 7 if self.cmd else 0) + self.cl,
                sum((
                    1 << 0 if command.is_config else 0,
                    1 << 1 if len(command.frame) == 24 else 0,
                    1 << 2 if command.is_query else 0
                )),
                1 << self.channel - 1
            )) + command.frame.pack,
            extended_id=False
        ))
        logging.debug(u"-> {}".format(command))

        self._request_event.wait(self.timeout)
        if self._last_command and self._last_command.is_query and self._response is None:
            self._response = command.response(frame.BackwardModuleError())
            logging.warning(f'      <- NOT ANSWERED {self._last_command}')
        self._last_command = None
        self._request_event.clear()

        if (isinstance(self._response,
                       BackwardModuleError) or self._response is None or self._response._value is None) and command.need_retry:
            logging.debug(f'RESEND {command}')
            self._response = self.send(command)

        return self._response


def str_bytes(_bytes: bytes): return ' '.join(hex(b)[2:].rjust(2, '0') for b in _bytes)
