import json
import logging
import os
import sys
import threading
import datetime

from django.db import models

from spread_core import mqtt
from spread_core.bam import subgineries
from spread_core.tools.settings_const import RESTART_TIMEOUT, CYCLES
from spread_dashboard import utils
from spread_dashboard.project import bam
from spread_dashboard.project.models import Location


class _IncBase(models.Model):
    commit_timer = None

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        super(_IncBase, self).save(*args, **kwargs)
        self.start_timer()

    def delete(self, *args, **kwargs):
        super(_IncBase, self).delete(*args, **kwargs)
        self.start_timer()

    @staticmethod
    def start_timer():
        if 'fill' in sys.argv:
            return
        if _IncBase.commit_timer:
            _IncBase.commit_timer.cancel()
            _IncBase.commit_timer = None

        _IncBase.commit_timer = threading.Timer(bam.settings[RESTART_TIMEOUT], function=_IncBase.on_timer)
        _IncBase.commit_timer.start()

    @staticmethod
    def on_timer():
        _IncBase.commit_timer = None
        _IncBase.dump()
        os.system(f'sudo supervisorctl restart spread_cycler')

    @staticmethod
    def dump():
        utils.write_file(
            file=bam.settings[CYCLES],
            data={
                'locations': [
                    utils.serialize(incubator)
                    for incubator in Incubator.objects.all()
                ],
                'cycles': [
                    utils.serialize(cycle)
                    for cycle in Cycle.objects.all()
                ],
                'triggers': [
                    {
                        **utils.serialize(trigger),
                        'topic': trigger.topic
                    }
                    for trigger in IncTrigger.objects.all()
                ]
            }
        )

    @staticmethod
    def fill():
        file = bam.configs.Dashboard().cycles
        ids = {Cycle: [], Incubator: [], IncTrigger: []}
        if os.path.exists(file):
            try:
                with open(file, 'r') as fs:
                    data = json.loads(fs.read())
            except BaseException as ex:
                logging.error(ex)
            else:
                for cycle_data in data.get('cycles', []):
                    ids[Cycle].append(
                        Cycle.objects.update_or_create(id=cycle_data['id'], defaults=cycle_data)[0].id
                    )
                for trigger_data in data.get('triggers', []):
                    trigger_data.pop('topic')
                    ids[IncTrigger].append(
                        IncTrigger.objects.update_or_create(id=trigger_data['id'], defaults=trigger_data)[0].id
                    )
                for loc_data in data.get('locations', []):
                    ids[Incubator].append(
                        Incubator.objects.update_or_create(id=loc_data['id'], defaults=loc_data)[0].id
                    )

                Incubator.objects.exclude(id__in=ids[Incubator]).delete()
                Cycle.objects.exclude(id__in=ids[Cycle]).delete()
                IncTrigger.objects.exclude(id__in=ids[IncTrigger]).delete()

    @staticmethod
    def clear_all():
        Incubator.objects.all().delete()
        Cycle.objects.all().delete()
        IncTrigger.objects.all().delete()


class Cycle(_IncBase):
    id = models.AutoField(primary_key=True)
    name = models.CharField(max_length=30, default='Новый цикл', verbose_name='Название цикла')
    period = models.PositiveIntegerField(default=0, verbose_name='Длительность (дней)')
    duration = models.PositiveIntegerField(default=50, verbose_name='Рассвет/закат (мин)')
    startLevel = models.PositiveIntegerField(default=25, verbose_name='Начальная яркость (%)')

    @staticmethod
    def copied_fields(): return Cycle.period.field.name, Cycle.duration.field.name, Cycle.startLevel.field.name

    def __str__(self): return self.name

    @property
    def triggers(self): return IncTrigger.objects.filter(cycle=self)


class Incubator(_IncBase):
    id = models.AutoField(primary_key=True)
    location = models.OneToOneField(Location, on_delete=models.CASCADE)
    cycle = models.ForeignKey(Cycle, null=True, blank=True, on_delete=models.SET_NULL, verbose_name='Цикл')
    startTime = models.DateField(default=datetime.date.today, max_length=16, verbose_name='Сегодня')
    active = models.BooleanField(default=False)
    recursive = models.BooleanField(default=False, verbose_name='Перезапускать')

    @property
    def current_day(self):
        if self.cycle.period == 0:
            return -1
        return (datetime.date.today() - self.startTime).days % self.cycle.period


class IncTrigger(_IncBase):
    commit_timer = None

    id = models.AutoField(primary_key=True)
    cycle = models.ForeignKey(Cycle, on_delete=models.CASCADE, editable=False)
    day = models.PositiveIntegerField()
    time = models.TimeField()
    level = models.PositiveIntegerField()

    @staticmethod
    def copied_fields(): return IncTrigger.day.field.name, IncTrigger.time.field.name, IncTrigger.level.field.name

    @property
    def topic(self):
        try:
            return str(mqtt.spread.topic.Set(
                mqtt.spread.address.SubgineryAddress(
                    bam.project.id,
                    bam.models.Subginery.objects.get(
                        locationID=Incubator.objects.get(cycle=self.cycle).location.id,
                        type=subgineries.Lighting.__name__
                    ).id,
                    subgineries.Lighting.__name__,
                    subgineries.BrightnessLevel
                )
            ))
        except BaseException as ex:
            logging.exception(ex)
            return None
