import json
import logging
import os
import uuid
import zipfile

import spread_core.mqtt as mqtt
from spread_core.mqtt.project import TopicProject
from spread_core.bam.generator import generate_provider
from spread_core.bam.dali.managers import Dali
from spread_core.bam.lom.manager import Lom
from spread_core.bam.mercury.managers import Mercury
from spread_dashboard.project import models
from . import configs


class Project:
    restarted = False
    changed = False
    inited = False

    def __init__(self):
        if os.path.exists(configs.Projector.ctp_file):
            ctp = zipfile.ZipFile(configs.Projector.ctp_file, 'r')
            self.id, self.rootLocation, self.name = self.read_pid(ctp)
            Project.inited = True

    @staticmethod
    def read_pid(ctp):
        entity = json.loads(ctp.read(TopicProject.HEADER + '.json').decode())[TopicProject.HEADER]
        return entity['project']['id'], entity['project']['rootLocationID'], entity['project']['name']

    @staticmethod
    def read_sid(ctp):
        for entity in json.loads(ctp.read(TopicProject.SERVERS + '.json').decode())[TopicProject.SERVERS]:
            if entity['type'] == 'Spread':
                return entity['id']

    @classmethod
    def read_ctp(cls):
        ctp = zipfile.ZipFile(configs.Projector.ctp_file, 'r')

        cls.read_ctp_file(models.Location, TopicProject.LOCATIONS, ctp)
        cls.read_ctp_file(models.Subginery, TopicProject.SUBGINERIES, ctp)
        cls.read_ctp_file(models.Manager, TopicProject.MANAGERS, ctp)
        cls.read_ctp_file(models.Enginery, TopicProject.ENGINERIES, ctp)
        cls.read_ctp_file(models.Provider, TopicProject.PROVIDERS, ctp)

    @staticmethod
    def read_ctp_file(model, file, ctp):
        stack = []
        for entity in json.loads(ctp.read(file + '.json').decode())[file]:
            try:
                assert issubclass(model, models.Entity)
                model.objects.update_or_create(id=entity['id'], defaults=model.separator(entity))
                stack.append(entity['id'])
            except BaseException as ex:
                logging.error('Filling error: %s' % ex)
        model.objects.exclude(id__in=stack).delete()

    @staticmethod
    def read_settings():
        sett = configs.Dashboard().serialize()
        for file in sett.pop('include', []):
            if os.path.exists(file):
                with open(file, 'r') as fs:
                    sett.update(json.loads(fs.read()))
        return sett

    @classmethod
    def generate_configs(cls):
        ctp = zipfile.ZipFile(configs.Projector.ctp_file, 'r')

        Project.changed = True
        configs.Config.PROJECT_ID, _, _ = cls.read_pid(ctp)
        configs.Frontier.SERVER_ID = cls.read_sid(ctp)

        tree = configs.Tree()

        configs.Common().save(generate_sv=True)
        configs.Frontier().save(generate_sv=True)
        configs.Dashboard().save(generate_sv=True)
        configs.Projector().save(generate_sv=True)
        configs.Scheduler().save(generate_sv=True)
        configs.Scripter().save(generate_sv=True)
        configs.Cycler().save(generate_sv=True)
        configs.EventsSqlite().save(generate_sv=True)
        configs.LightingEquipment().save(generate_sv=True)

        adapters = []
        for manager in json.loads(ctp.read(TopicProject.MANAGERS + '.json').decode())[TopicProject.MANAGERS]:
            adapter = None
            if manager['type'] == Dali.__name__:
                bus = manager['attributes']['bus']
                m = configs.DaliManager(bus)
                if bus not in adapters:
                    adapters.append(bus)
                    adapter = configs.CanAdapter(bus)
            elif manager['type'] == Mercury.__name__:
                bus = manager['attributes']['bus']
                m = configs.MercuryManager(bus)
                if bus not in adapters:
                    adapters.append(bus)
                    adapter = configs.TcpAdapter(bus)
            elif manager['type'] == Lom.__name__:
                m = configs.LomManager()
            else:
                continue

            if adapter:
                adapter.save(generate_sv=True)
                tree.decrease(adapter.config_dest, adapter.file)
            m.managerID = manager['id']
            m.save(generate_sv=True)
            tree.decrease(m.config_dest, m.file)

        tree.clear()


project = Project()
if project.inited:
    configs.Config.PROJECT_ID = project.id
settings = Project.read_settings()


def broker_auth() -> dict:
    guid = str(uuid.uuid4())
    reply = str(mqtt.spread.topic.Reply(project.id, guid))
    return dict(web_port=2883, login='admin', password='broker', reply=reply, guid=guid)


def config(model: models.Provider):
    guid = uuid.uuid4()
    provider = generate_provider(model.serialize())
    manager_model = models.Manager.objects.get(id=model.managerID)
    manager = (Dali if manager_model.type == Dali.__name__ else Mercury)(project.id, manager_model.id)
    manager.parse(manager_model.serialize())
    provider.set_manager(manager)
    return dict(broker=broker_auth(),
                project=dict(state=str(mqtt.spread.topic.State.of_provider(provider, '+')),
                             get=str(mqtt.spread.topic.Renew.of_provider(provider, '{}')),
                             set=str(mqtt.spread.topic.Set.of_provider(provider, '{}')),
                             reply=str(mqtt.spread.topic.Reply(project.id, guid)),
                             session_id=str(guid),
                             entity_id=model.id))
