import json
import logging
import os
import re

import math
from django.db import models
import jsonfield

from . import fields

CONFIG_PATH = '/etc/spread/'
LOG_PATH = '/var/log/spread/'
DUMP_PATH = '/var/project/'
SV_PATH = '/etc/supervisor/conf.d/'
SV_TEXT = """[program:enter_name]
command=daphne -u /run/spread_dashboard.sock --fd 0 --access-log - --proxy-headers spread_dashboard.asgi:application
;process_name=%(program_name)s_%(process_num)02d
numprocs=1
directory=/var/spread
;umask=022
;priority=999
autostart=true
autorestart=true
startsecs=3
startretries=3
;exitcodes=0,2
;stopsignal=TERM
stopsignal=QUIT
stopwaitsecs=10
;stopasgroup=false
;killasgroup=false
user=root
redirect_stderr=false
stdout_logfile=/var/log/spread/dashboard_stdout.log
stdout_logfile_maxbytes=3MB
stdout_logfile_backups=10
;stdout_capture_maxbytes=1MB
;stdout_events_enabled=false
stderr_logfile=/var/log/spread/dashboard_stderr.log
stderr_logfile_maxbytes=1MB
stderr_logfile_backups=10
;stderr_capture_maxbytes=1MB
;stderr_events_enabled=false
;environment=A="1",B="2"
;serverurl=AUTO
"""


class Tree(dict):
    def __init__(self):
        super(Tree, self).__init__(**{
            root.split('/')[-1]: sorted([file.split('.')[0] for file in files])
            for root, dicts, files in os.walk(os.path.join(CONFIG_PATH))
        })

    def decrease(self, destination, name):
        if destination in self:
            if name in self[destination]:
                self[destination].remove(name)

    def clear(self) -> None:
        for dest in ('manager', 'adapter'):
            if dest in self:
                for m in self[dest]:
                    conf_path = os.path.join(CONFIG_PATH, dest, m+'.json')
                    if os.path.exists(conf_path):
                        os.remove(conf_path)
                    sv_path = os.path.join(SV_PATH, ('spread_%s_%s.conf' % (dest, m)).replace('__', '_'))
                    if os.path.exists(sv_path):
                        os.remove(sv_path)
                self.pop(dest)


class Config(models.Model):
    PROJECT_ID = 0
    config_dest = ''
    file = None
    sv_program = 'program'
    sv_key_autorestart = 'true'
    sv_key_startsec = 3

    class Meta:
        abstract = True

    @property
    def sv_key_command(self):
        res = 'spread_%s_%s --config %s' % (self.config_dest, re.sub(r'(_\d+)*$', '', self.file), self.config_file)
        return res.replace('__', '_')

    @property
    def sv_key_stderr_logfile(self): return os.path.join(LOG_PATH, self.config_dest, self.file + '_stderr.log')

    @property
    def sv_key_stdout_logfile(self): return os.path.join(LOG_PATH, self.config_dest, self.file + '_stdout.log')

    @property
    def sv_data(self):
        s = re.sub(r'.+program:.+', '[%s:%s]' % (self.sv_program, self.service_name), SV_TEXT)

        for var_name in [key for key in dir(self) if key.startswith('sv_key_')]:
            key = re.sub(r'sv_key_', '', var_name)
            s = re.sub(r';?%s=.+' % key, '%s=%s' % (key, getattr(self, var_name)), s)

        return s

    @property
    def service_name(self):
        return ('spread_%s_%s' % (self.config_dest, self.file)).replace('__', '_')

    @property
    def config_file(self):
        return os.path.join(CONFIG_PATH, self.config_dest, self.file + '.json')

    @property
    def dump_file(self):
        return os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'dump', self.config_dest, self.file + '.json')

    @property
    def sv_file(self):
        return os.path.join(SV_PATH, self.service_name + '.conf')

    def __init__(self, *args, **kwargs):
        super(Config, self).__init__(*args, **{**self.read_file(self.config_file), **kwargs})
        if Config.PROJECT_ID:
            for field in self._meta.fields:
                if hasattr(self, field.attname):
                    attr_value = getattr(self, field.attname)
                    if isinstance(attr_value, str):
                        _p_id = re.findall(r'^%s(\d+)/.*$' % DUMP_PATH, attr_value)
                        if _p_id and int(_p_id[0]) != str(Config.PROJECT_ID):
                            setattr(self, field.attname, '')

        if hasattr(self, 'include'):
            self.include = [os.path.join(CONFIG_PATH, 'common.json')]
        if hasattr(self, 'dump') and not self.dump:
            self.dump = self.dump_file

    def serialize(self) -> dict:
        return {field.attname: getattr(self, field.attname) for field in self._meta.fields
                if field.serialize and
                not (field.attname == 'logLevel' and not getattr(self, 'logLevel'))}

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None, generate_sv=False):
        self.write_file(self.config_file, self.serialize())
        self.check_sv_log()
        if generate_sv:
            self.save_sv()

    def check_sv_log(self):
        if self.sv_file:
            os.makedirs(os.path.dirname(self.sv_key_stdout_logfile), exist_ok=True)
            os.makedirs(os.path.dirname(self.sv_key_stderr_logfile), exist_ok=True)

    def save_sv(self):
        if self.sv_file:
            self.write_file(self.sv_file, self.sv_data)

    @staticmethod
    def read_file(path):
        try:
            with open(path, 'r') as file:
                return json.loads(file.read())
        except:
            return {}

    @staticmethod
    def write_file(path, data):
        logging.info('save %s' % path)
        if not isinstance(data, (str, int)):
            data = json.dumps(data, indent=2)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as file:
            file.write(data)


class Common(Config):
    class Meta:
        abstract = True
    file = 'common'
    sv_file = None

    projectID = models.IntegerField(verbose_name='Идентификатор проекта')
    logLevel = fields.LogLevelCommonField()
    host = models.GenericIPAddressField(default='127.0.0.1', verbose_name='Адрес брокера')
    port = models.PositiveIntegerField(default=1883, verbose_name='Прослушиваемый порт')
    keepalive = models.PositiveIntegerField(default=60, verbose_name='Время жизни соединения')
    login = models.CharField(max_length=255, default='admin', verbose_name='Логин')
    password = models.CharField(max_length=255, default='broker', verbose_name='Пароль')

    broker_fields = [host, port, keepalive, login, password]

    def __init__(self, *args, **kwargs):
        super(Common, self).__init__(*args, **kwargs)
        self.projectID = Config.PROJECT_ID

    def serialize(self) -> dict:
        data = super(Common, self).serialize()
        return {'broker': {key: data.pop(key) for key in [field.attname for field in self.broker_fields]}, **data}

    @staticmethod
    def read_file(path):
        data = Config.read_file(path)
        return {**data.pop('broker', {}), **data}


class ConfigOfManager(Config):
    class Meta:
        abstract = True

    config_dest = 'manager'

    include = fields.IncludeField()
    managerID = models.IntegerField(verbose_name='Идентификатор')
    logLevel = fields.LogLevelField()

    def __init__(self, bus, *args, **kwargs):
        self.index = '_'.join(re.findall(r'\d+', bus))
        super(ConfigOfManager, self).__init__(*args, **kwargs)


class DaliManager(ConfigOfManager):
    file = property(fget=lambda self: 'dali_%s' % self.index)

    dump = fields.DumpField()
    communicateTimeout = models.FloatField(default=1.0, verbose_name='Время ожидания ответа')
    localAdapter = models.BooleanField(default=False, verbose_name='Локальный can-адаптер')


class CanAdapter(Config):
    config_dest = 'adapter'
    file = property(fget=lambda self: 'rbus_%s' % self.index)

    include = fields.IncludeField()
    canInterface = models.CharField(max_length=7, verbose_name='Название Can-интерфейса')
    subtopicCan = models.CharField(max_length=25, verbose_name='Can-топик')
    subtopicDali = models.CharField(max_length=25, verbose_name='Dali-топик')
    subtopicModbus = models.CharField(max_length=25, verbose_name='Modbus-топик')
    canReopenTimeout = models.PositiveIntegerField(default=1000, help_text='Время в миллисекундах',
                                                   verbose_name='Интервал переоткрытия интерфейса')
    logLevel = fields.LogLevel2Field()
    enableSysLog = models.BooleanField(default=False, verbose_name='Системные логи')
    enableCanSendTopic = models.BooleanField(default=False, verbose_name='Трансляция в CanSend-топик')
    enableCanDumpTopic = models.BooleanField(default=False, verbose_name='Трансляция в CanDump-топик')
    appendTimestamp = models.BooleanField(default=True, verbose_name='Указывать Timestamp')

    def __init__(self, index, *args, **kwargs):
        self.index = int(re.findall(r'\d+', index)[0])
        self.interface = ('' if self.index == 0 else 'r') + 'can%s' % self.index
        super(CanAdapter, self).__init__(*args, **{
            'canInterface': self.interface,
            'subtopicCan': 'Rapida/%s/Can' % self.index,
            'subtopicDali': 'Rapida/%s/Dali' % self.index,
            'subtopicModbus': 'Rapida/%s/Modbus' % self.index,
            **kwargs
        })


class MercuryManager(ConfigOfManager):
    file = property(fget=lambda self: 'mercury_%s' % self.index)

    communicateTimeout = models.PositiveIntegerField(default=1, verbose_name='Время ожидания ответа')
    hexed = jsonfield.JSONField(default=[], blank=True, verbose_name='Адреса счётчиков с индексом "D"')

    def clean_fields(self, exclude=None):
        if exclude is None: exclude = []
        exclude.append('hexed')
        super(MercuryManager, self).clean_fields(exclude)
        if self.hexed is None:
            self.hexed = []
        elif not isinstance(self.hexed, list):
            self.hexed = [self.hexed]


class TcpAdapter(Config):
    config_dest = 'adapter'
    file = property(fget=lambda self: 'tcp_%s' % self.index)

    include = fields.IncludeField()

    subtopic = models.CharField(max_length=25)
    # "10.10.1.32"
    host = models.GenericIPAddressField(default='', verbose_name='Адрес Moxa-адаптера')
    port = models.PositiveIntegerField(default=4001, verbose_name='Порт Moxa-адаптера')
    timeout = models.FloatField(default=0.7, verbose_name='Время ожидания ответа')
    killTimeout = models.FloatField(default=3, verbose_name='Время простоя TCP-соединения')
    logLevel = fields.LogLevelField()

    def __init__(self, bus, *args, **kwargs):
        self.index = int(re.findall(r'\d+', bus)[0])
        super(TcpAdapter, self).__init__(*args, **kwargs)
        if not self.subtopic:
            self.subtopic = bus

    @staticmethod
    def read_file(path):
        data = Config.read_file(path)
        return {**data.pop('bus', {}), **data}

    def serialize(self) -> dict:
        data = super(TcpAdapter, self).serialize()
        return {'include': data.pop('include'), 'bus': {**data}}


class LomManager(ConfigOfManager):
    file = 'lom'

    def __init__(self, *args, **kwargs):
        super(LomManager, self).__init__('1', *args, **kwargs)
        if not self.profiles:
            self.profiles = self.profiles_file

    @property
    def profiles_file(self):
        return os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'dump', self.config_dest, 'lom_profiles.json')

    dump = fields.DumpField()
    profiles = models.CharField(max_length=255, verbose_name='Файл профилей')
    tempDimmingPeriod = models.PositiveIntegerField(default=30, verbose_name='Длительность временного диммирования')


class Frontier(Config):
    SERVER_ID = 0
    file = 'frontier'

    include = fields.IncludeField()
    serverID = models.IntegerField()
    logLevel = fields.LogLevelField()

    def __init__(self, *args, **kwargs):
        super(Frontier, self).__init__(*args, **kwargs)
        self.serverID = self.SERVER_ID


class Scheduler(Config):
    file = 'scheduler'
    calendar_file = 'calendar'

    include = fields.IncludeField()
    tasks = models.CharField(max_length=255)
    calendar = models.CharField(max_length=255)
    logLevel = fields.LogLevelField()

    sv_key_autorestart = 'unexpected'
    sv_key_exitcodes = 0
    sv_key_startsec = 0

    def __init__(self, *args, **kwargs):
        super(Scheduler, self).__init__(*args, **kwargs)
        if Config.PROJECT_ID:
            if not self.tasks:
                self.tasks = os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'dump', 'handling', self.file+'.json')
            if not self.calendar:
                self.calendar = os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'dump', 'handling', self.calendar_file+'.json')


class Cycler(Config):
    file = 'cycler'
    FADE_TIMES = ((round(math.sqrt(2**i)/2, 1), round(math.sqrt(2**i)/2, 1)) for i in range(1, 16))

    include = fields.IncludeField()
    data = models.CharField(max_length=255)
    logLevel = fields.LogLevelField()
    fadeTime = models.FloatField(choices=FADE_TIMES, default=90.5)

    sv_key_autorestart = 'unexpected'
    sv_key_exitcodes = 0
    sv_key_startsec = 0

    def __init__(self, *args, **kwargs):
        super(Cycler, self).__init__(*args, **kwargs)
        if Config.PROJECT_ID:
            if not self.data:
                self.data = os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'dump', 'handling', self.file+'.json')


class Scripter(Config):
    file = 'scripter'

    sv_key_autorestart = 'unexpected'
    sv_key_exitcodes = 0
    sv_key_startsec = 0

    include = fields.IncludeField()
    triggers = models.CharField(max_length=255)
    calendar = models.CharField(max_length=255)
    logLevel = fields.LogLevelField()

    def __init__(self, *args, **kwargs):
        super(Scripter, self).__init__(*args, **kwargs)
        if Config.PROJECT_ID:
            if not self.triggers:
                self.triggers = os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'dump', 'handling', self.file+'.json')
            if not self.calendar:
                self.calendar = os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'dump', 'handling', Scheduler.calendar_file+'.json')


class Projector(Config):
    file = 'projector'
    ctp_file = '/var/project/throne-project.ctp'
    mosquitto_file = '/var/lib/mosquitto/mosquitto.db'

    include = fields.IncludeField()
    ctpPath = models.CharField(max_length=255, default=ctp_file)
    localPath = models.CharField(max_length=255, default='')
    mosquittoDB = models.CharField(max_length=255, default=mosquitto_file)
    logLevel = fields.LogLevelField()

    def __init__(self, *args, **kwargs):
        super(Projector, self).__init__(*args, **kwargs)
        if Config.PROJECT_ID:
            if not self.localPath:
                self.localPath = os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'tass')


class EventsSqlite(Config):
    config_dest = 'db'
    file = 'events_sqlite'

    include = fields.IncludeField()
    database = models.CharField(max_length=255)
    enableSysLog = models.BooleanField(default=False)
    tickRate = models.PositiveIntegerField(default=60000)
    failureOffset = models.PositiveIntegerField(default=1000)
    sizeLimit = models.PositiveIntegerField(default=256)
    logLevel = fields.LogLevel2Field()

    def __init__(self, *args, **kwargs):
        super(EventsSqlite, self).__init__(*args, **kwargs)
        if Config.PROJECT_ID:
            if not self.database:
                self.database = os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'tass', 'tass.sqlite')


class LightingEquipment(Config):
    config_dest = 'equipment'
    file = 'lighting'

    include = fields.IncludeField()
    dump = fields.DumpField()
    logLevel = fields.LogLevelField()


class Dashboard(Config):
    file = 'dashboard'

    sv_program = 'fcgi-program'
    sv_key_command = 'daphne -u /run/spread_dashboard.sock --fd 0 --access-log - --proxy-headers spread_dashboard.asgi:application'

    include = fields.IncludeField()
    port = models.PositiveIntegerField(serialize=False, default=55582, verbose_name='nginx-порт')
    httpRoot = models.CharField(max_length=255, default='spread')
    restartTimeout = models.PositiveIntegerField(default=10)
    tasks = models.CharField(max_length=255)
    cycles = models.CharField(max_length=255, default='')
    calendar = models.CharField(max_length=255)
    logLevel = fields.LogLevelField()

    def __init__(self, *args, **kwargs):
        super(Dashboard, self).__init__(*args, **kwargs)
        if Config.PROJECT_ID:
            if not self.tasks:
                self.tasks = os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'dump', 'handling', Scheduler.file+'.json')
            if not self.calendar:
                self.calendar = os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'dump', 'handling', Scheduler.calendar_file+'.json')
            if not self.cycles:
                self.cycles = os.path.join(DUMP_PATH, str(Config.PROJECT_ID), 'dump', 'handling', Cycler.file + '.json')

    @property
    def sv_data(self):
        s = super(Dashboard, self).sv_data
        s += 'socket=tcp://localhost:' + str(self.port) + '\n;socket_owner=root\n;socket_mode=0700\n'
        s = re.sub(r';?environment=.+', 'environment=%s' % 'DJANGO_SETTINGS_MODULE=spread_dashboard.settings', s)

        return s
