import json
import logging
import os
import sys
import threading

from django.db import models
from jsonfield import JSONField

import spread_core.mqtt as mqtt
from spread_core.tools.settings_const import RESTART_TIMEOUT, TASKS, CALENDAR
from spread_core.bam import engineries, handling, subgineries
from spread_dashboard import utils
from spread_dashboard.project import bam

COMMANDS = {
    subgineries.Lighting.__name__: dict(
        name='Вся локация',
        topic=lambda e_id, funit: mqtt.spread.topic.Set(mqtt.spread.address.SubgineryAddress(bam.project.id, e_id, subgineries.Lighting.__name__, funit)),
        commands=[
            dict(values=[True], name='Включить всё', funit=subgineries.On.funit_type),
            dict(values=[1], name='Световая сцена I', funit=subgineries.SceneOn.funit_type),
            dict(values=[2], name='Световая сцена II', funit=subgineries.SceneOn.funit_type),
            dict(values=[True], name='Выключить всё', funit=subgineries.Off.funit_type),
            dict(validator=dict(min=0, max=100), name='Яркость', funit=subgineries.BrightnessLevel.funit_type),
        ]
    ),
    engineries.LightingArea.__name__: dict(
        topic=lambda e_id, funit: mqtt.spread.topic.Set(mqtt.spread.address.EngineryAddress(bam.project.id, engineries.LightingArea.__name__, e_id, funit)),
        commands=[
            dict(name='Выключить', funit=engineries.lighting_area.On.funit_type, values=[False]),
            dict(name='Включить', funit=engineries.lighting_area.On.funit_type, values=[True]),
            dict(name='Целевая освещённость', funit=engineries.lighting_area.TargetLuminosity.funit_type),
            dict(name='Реакция на движение', funit=engineries.lighting_area.OccupancyAction.funit_type, values=['qwe', 2]),
        ]
    ),
    engineries.DimmingLight.__name__: dict(
        name='Светильник 1',
        topic=lambda e_id, funit: mqtt.spread.topic.Set(mqtt.spread.address.EngineryAddress(bam.project.id, engineries.DimmingLight.__name__, e_id, funit)),
        entity=999,
        commands=[
            dict(name='Включить', funit=engineries.light.On.funit_type, values=[True]),
            dict(name='Выключить', funit=engineries.light.On.funit_type, values=[False]),
            dict(name='Яркость', funit=engineries.light.BrightnessLevel.funit_type, validator=dict(min=0, max=100)),
            dict(name='Групповая яркость', funit=engineries.light.BrightnessLevel.funit_type, validator=dict(min=0, max=100)),
        ]
    ),
    engineries.TunableWhiteLight.__name__: dict(
        name='TW 1',
        topic=lambda e_id, funit: mqtt.spread.topic.Set(mqtt.spread.address.EngineryAddress(bam.project.id, engineries.TunableWhiteLight.__class__.__name__, e_id, funit)),
        entity=999,
        commands=[
            dict(name='Включить', funit=engineries.light.On.funit_type, values=[True]),
            dict(name='Выключить', funit=engineries.light.On.funit_type, values=[False]),
            dict(name='Яркость', funit=engineries.light.BrightnessLevel.funit_type, validator=dict(min=0, max=100)),
            dict(name='Групповая яркость', funit=engineries.light.GroupLevel.funit_type, validator=dict(min=0, max=100)),
            dict(name='Температура', funit=engineries.light.Temperature.funit_type, validator=dict(min=2000, max=8000)),
        ]
    )
}


class SchBase(models.Model):
    commit_timer = None

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        super(SchBase, self).save(*args, **kwargs)
        self.start_timer()

    def delete(self, *args, **kwargs):
        super(SchBase, self).delete(*args, **kwargs)
        self.start_timer()

    @staticmethod
    def start_timer():
        if 'fill' in sys.argv:
            return
        if SchBase.commit_timer:
            SchBase.commit_timer.cancel()
            SchBase.commit_timer = None

        SchBase.commit_timer = threading.Timer(bam.settings[RESTART_TIMEOUT], function=SchBase.on_timer)
        SchBase.commit_timer.start()

    @staticmethod
    def on_timer():
        SchBase.commit_timer = None
        SchBase.dump()
        Holiday.dump()
        os.system(f'sudo supervisorctl restart spread_scheduler')

    @staticmethod
    def dump():
        utils.write_file(
            bam.settings[TASKS],
            [{
                **utils.serialize(location),
                'triggers': [
                    {
                        **utils.serialize(trigger),
                        'actions': [
                            utils.serialize(action) for action in SchAction.objects.filter(trigger=trigger)
                        ]
                    } for trigger in SchTrigger.objects.filter(location=location)
                ]
            } for location in SchLocation.objects.all()]
        )

    @staticmethod
    def fill():
        file = bam.configs.Dashboard().tasks
        if os.path.exists(file):
            try:
                with open(file, 'r') as fs:
                    data = json.loads(fs.read())
            except BaseException as ex:
                logging.error(ex)
            else:
                for _loc in data:
                    triggers_data = _loc.pop('triggers')
                    location, _ = SchLocation.objects.update_or_create(id=_loc['id'], defaults=_loc)
                    try:
                        bam.models.Location.objects.get(
                            scheduled=True,
                            id=bam.models.Subginery.objects.get(
                                id=location.id
                            ).locationID
                        )
                    except (bam.models.Location.DoesNotExist, bam.models.Subginery.DoesNotExist):
                        location.delete()
                    else:
                        for _tr in triggers_data:
                            actions_data = _tr.pop('actions')
                            trigger, _ = SchTrigger.objects.update_or_create(id=_tr['id'],
                                                                             defaults={**_tr, 'location': location})
                            for _act in actions_data:
                                SchAction.objects.update_or_create(id=_act['id'], defaults={**_act, 'trigger': trigger})

                SchLocation.objects.filter(
                    id__in=[subginery.id for subginery in (bam.models.Subginery.objects.filter(
                        locationID__in=[location.id for location in bam.models.Location.objects.filter(scheduled=False)]
                    ))]
                ).delete()

        for loc in bam.models.Location.objects.filter(scheduled=True):
            try:
                SchLocation.objects.update_or_create(
                    id=bam.models.Subginery.objects.get(locationID=loc.id, type=subgineries.Lighting.__name__).id,
                    defaults={'name': loc.name}
                )
            except:
                continue


class SchLocation(SchBase):
    id = models.AutoField(primary_key=True)
    enabled = models.BooleanField(default=False)
    name = models.CharField(max_length=255)


class SchTrigger(SchBase):
    id = models.AutoField(primary_key=True)
    type = models.CharField(max_length=255)#, choices=[('workday', 'Рабочий день'), ('holiday', 'Выходной')])
    location = models.ForeignKey(SchLocation, on_delete=models.CASCADE, serialize=False)
    time = models.TimeField()


class SchAction(SchBase):
    id = models.AutoField(primary_key=True)
    trigger = models.ForeignKey(SchTrigger, on_delete=models.CASCADE, serialize=False)
    type = models.CharField(max_length=255, default='broker', editable=False)
    topic = models.CharField(max_length=255)
    value = JSONField()

    def to_view_data(self):
        action = handling.actions.action_of_data(utils.serialize(self))
        assert isinstance(action, handling.BrokerAction)
        target = action.entity_address.type

        commands = COMMANDS[target]['commands']
        for i in range(0, len(commands)):
            funit = action.entity_address.funit_type
            if commands[i]['funit'] == funit:
                value = action.value
                if 'values' in commands[i]:
                    if len(commands[i]['values']) == 1:
                        if commands[i]['values'][0] != value:
                            continue
                    elif len(commands[i]['values']) > 1:
                        value = commands[i]['values'].index(value)
                return dict(
                    id=action.id,
                    target=list(COMMANDS.keys()).index(target),
                    action=i,
                    value=value
                )
        return {}

    @staticmethod
    def from_view_data(loc_id, tr_id, target, act_id, index, value):
        e_type = list(COMMANDS.keys())[target]
        item = COMMANDS[e_type]['commands'][index]
        topic = COMMANDS[e_type]['topic'](loc_id, item['funit'])

        if 'values' not in item:
            value = json.loads(value)
        elif len(item['values']) == 1:
            value = item['values'][0]
        else:
            value = item['values'][int(value)]

        if act_id:
            model = SchAction.objects.get(id=act_id)
            model.topic = topic
            model.value = value
            model.save()
        else:
            trigger = SchTrigger.objects.get(id=tr_id)
            model = SchAction.objects.create(**dict(trigger=trigger, topic=topic, value=value))

        return model

    @staticmethod
    def actions(loc_id: int):
        return [
            {**COMMANDS[subgineries.Lighting.__name__], 'entity': loc_id},
            *[
                {
                    **COMMANDS[e.type],
                    'name': e.name,
                    'entity': e.id
                } for e in bam.models.Enginery.get_scheduled(loc_id)
            ]
        ]


class Holiday(SchBase):
    date = models.DateField(unique=True)

    @staticmethod
    def dump():
        utils.write_file(
            bam.settings[CALENDAR],
            [str(model.date) for model in Holiday.objects.all()]
        )

    @staticmethod
    def fill():
        file = bam.configs.Dashboard().calendar
        if os.path.exists(file):
            try:
                with open(file, 'r') as fs:
                    data = json.loads(fs.read())
            except BaseException as ex:
                logging.error(ex)
            else:
                Holiday.objects.all().delete()
                for datum in data:
                    Holiday.objects.create(date=datum)
